from calendar import timegm
from datetime import date, datetime
from decimal import Decimal
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

from django.core import mail
from django.test import override_settings, tag
from django.utils.timezone import (
    get_fixed_timezone,
    override as override_current_timezone,
)

from anymail.exceptions import (
    AnymailAPIError,
    AnymailConfigurationError,
    AnymailRecipientsRefused,
    AnymailSerializationError,
    AnymailUnsupportedFeature,
)
from anymail.message import attach_inline_image_file

from .mock_requests_backend import RequestsBackendMockAPITestCase
from .utils import (
    SAMPLE_IMAGE_FILENAME,
    decode_att,
    sample_image_content,
    sample_image_path,
)


@tag("mailersend")
@override_settings(
    EMAIL_BACKEND="anymail.backends.mailersend.EmailBackend",
    ANYMAIL={"MAILERSEND_API_TOKEN": "test_api_token"},
)
class MailerSendBackendMockAPITestCase(RequestsBackendMockAPITestCase):
    """TestCase that uses MailerSend EmailBackend with a mocked API"""

    DEFAULT_STATUS_CODE = 202
    DEFAULT_RAW_RESPONSE = b""
    DEFAULT_CONTENT_TYPE = "text/html"

    def setUp(self):
        super().setUp()
        # Simple message useful for many tests
        self.message = mail.EmailMultiAlternatives(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )

    def set_mock_success(self, message_id="1234567890abcdef"):
        response = self.set_mock_response()
        if message_id is not None:
            response.headers["x-message-id"] = message_id
        return response

    def set_mock_rejected(
        self, rejections, warning_type="ALL_SUPPRESSED", message_id="1234567890abcdef"
    ):
        """rejections should be a dict of {email: [reject_reason, ...], ...}"""
        if warning_type == "ALL_SUPPRESSED":
            message_id = None
        response = self.set_mock_response(
            json_data={
                "warnings": [
                    {
                        "type": warning_type,
                        "recipients": [
                            {"email": email, "reasons": reasons}
                            for email, reasons in rejections.items()
                        ],
                    }
                ]
            }
        )
        if message_id is not None:
            response.headers["x-message-id"] = message_id
        return response


@tag("mailersend")
class MailerSendBackendStandardEmailTests(MailerSendBackendMockAPITestCase):
    """Test backend support for Django standard email features"""

    def test_send_mail(self):
        """Test basic API for simple send"""
        mail.send_mail(
            "Subject here",
            "Here is the message.",
            "from@example.com",
            ["to@example.com"],
            fail_silently=False,
        )

        self.assert_esp_called("/v1/email")

        headers = self.get_api_call_headers()
        self.assertEqual(headers["Authorization"], "Bearer test_api_token")

        data = self.get_api_call_json()
        self.assertEqual(data["subject"], "Subject here")
        self.assertEqual(data["text"], "Here is the message.")
        self.assertEqual(data["from"], {"email": "from@example.com"})
        self.assertEqual(data["to"], [{"email": "to@example.com"}])

    def test_name_addr(self):
        """Make sure RFC2822 name-addr format (with display-name) is allowed

        (Test both sender and recipient addresses)
        """
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "From Name <from@example.com>",
            ["Recipient #1 <to1@example.com>", "to2@example.com"],
            cc=["Carbon Copy <cc1@example.com>", "cc2@example.com"],
            bcc=["Blind Copy <bcc1@example.com>", "bcc2@example.com"],
        )
        msg.send()

        data = self.get_api_call_json()
        self.assertEqual(
            data["from"], {"email": "from@example.com", "name": "From Name"}
        )
        self.assertEqual(
            data["to"],
            [
                {"email": "to1@example.com", "name": "Recipient #1"},
                {"email": "to2@example.com"},
            ],
        )
        self.assertEqual(
            data["cc"],
            [
                {"email": "cc1@example.com", "name": "Carbon Copy"},
                {"email": "cc2@example.com"},
            ],
        )
        self.assertEqual(
            data["bcc"],
            [
                {"email": "bcc1@example.com", "name": "Blind Copy"},
                {"email": "bcc2@example.com"},
            ],
        )

    def test_custom_headers(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com"],
            headers={
                "Reply-To": "another@example.com",
                "In-Reply-To": "12345@example.com",
                "Precedence": "Bulk",
                # Other custom headers only available to enterprise accounts:
                "X-Custom": "custom header",
            },
        )
        email.send()
        data = self.get_api_call_json()
        # Special handling headers:
        self.assertEqual(data["reply_to"], {"email": "another@example.com"})
        self.assertEqual(data["in_reply_to"], "12345@example.com")
        self.assertIs(data["precedence_bulk"], True)
        # Other headers:
        self.assertEqual(
            data["headers"], [{"name": "X-Custom", "value": "custom header"}]
        )

    def test_html_message(self):
        text_content = "This is an important message."
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMultiAlternatives(
            "Subject", text_content, "from@example.com", ["to@example.com"]
        )
        email.attach_alternative(html_content, "text/html")
        email.send()

        data = self.get_api_call_json()
        self.assertEqual(data["text"], text_content)
        self.assertEqual(data["html"], html_content)
        # Don't accidentally send the html part as an attachment:
        self.assertNotIn("attachments", data)

    def test_html_only_message(self):
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMessage(
            "Subject", html_content, "from@example.com", ["to@example.com"]
        )
        email.content_subtype = "html"  # Main content is now text/html
        email.send()

        data = self.get_api_call_json()
        self.assertNotIn("text", data)
        self.assertEqual(data["html"], html_content)

    def test_reply_to(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com"],
            reply_to=["Reply Name <reply@example.com>"],
        )
        email.send()

        data = self.get_api_call_json()
        self.assertEqual(
            data["reply_to"], {"email": "reply@example.com", "name": "Reply Name"}
        )

    def test_attachments(self):
        text_content = "* Item one\n* Item two\n* Item three"
        self.message.attach(
            filename="test.txt", content=text_content, mimetype="text/plain"
        )

        # Should guess mimetype if not provided...
        png_content = b"PNG\xb4 pretend this is the contents of a png file"
        self.message.attach(filename="test.png", content=png_content)

        # Should work with a MIMEBase object (also tests no filename)...
        pdf_content = b"PDF\xb4 pretend this is valid pdf params"
        mimeattachment = MIMEBase("application", "pdf")
        mimeattachment.set_payload(pdf_content)
        self.message.attach(mimeattachment)

        self.message.send()
        data = self.get_api_call_json()
        attachments = data["attachments"]
        self.assertEqual(len(attachments), 3)
        self.assertEqual(attachments[0]["disposition"], "attachment")
        self.assertEqual(attachments[0]["filename"], "test.txt")
        self.assertEqual(
            decode_att(attachments[0]["content"]).decode("ascii"), text_content
        )
        self.assertEqual(attachments[1]["disposition"], "attachment")
        self.assertEqual(attachments[1]["filename"], "test.png")
        self.assertEqual(decode_att(attachments[1]["content"]), png_content)
        self.assertEqual(attachments[2]["disposition"], "attachment")
        self.assertEqual(attachments[2]["filename"], "attachment.pdf")  # generated
        self.assertEqual(decode_att(attachments[2]["content"]), pdf_content)

    def test_unicode_attachment_correctly_decoded(self):
        # Slight modification from the Django unicode docs:
        # http://django.readthedocs.org/en/latest/ref/unicode.html#email
        self.message.attach(
            "Une pièce jointe.html", "<p>\u2019</p>", mimetype="text/html"
        )
        self.message.send()
        data = self.get_api_call_json()
        attachments = data["attachments"]
        self.assertEqual(len(attachments), 1)

    def test_embedded_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        cid = attach_inline_image_file(self.message, image_path)
        html_content = (
            '<p>This has an <img src="cid:%s" alt="inline" /> image.</p>' % cid
        )
        self.message.attach_alternative(html_content, "text/html")

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["html"], html_content)

        self.assertEqual(len(data["attachments"]), 1)
        self.assertEqual(data["attachments"][0]["disposition"], "inline")
        self.assertEqual(data["attachments"][0]["filename"], image_filename)
        self.assertEqual(data["attachments"][0]["id"], cid)
        self.assertEqual(decode_att(data["attachments"][0]["content"]), image_data)

    def test_attached_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        # option 1: attach as a file
        self.message.attach_file(image_path)

        # option 2: construct the MIMEImage and attach it directly
        image = MIMEImage(image_data)
        self.message.attach(image)

        self.message.send()
        data = self.get_api_call_json()
        attachments = data["attachments"]
        self.assertEqual(len(attachments), 2)
        self.assertEqual(attachments[0]["disposition"], "attachment")
        self.assertEqual(attachments[0]["filename"], image_filename)
        self.assertEqual(decode_att(attachments[0]["content"]), image_data)
        self.assertNotIn("id", attachments[0])  # not inline
        self.assertEqual(attachments[1]["disposition"], "attachment")
        self.assertEqual(attachments[1]["filename"], "attachment.png")  # generated
        self.assertEqual(decode_att(attachments[1]["content"]), image_data)
        self.assertNotIn("id", attachments[0])  # not inline

    def test_multiple_html_alternatives(self):
        # Multiple text/html alternatives not allowed
        self.message.attach_alternative("<p>First html is OK</p>", "text/html")
        self.message.attach_alternative("<p>But not second html</p>", "text/html")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_html_alternative(self):
        # Only html alternatives allowed
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_alternatives_fail_silently(self):
        # Make sure fail_silently is respected
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        sent = self.message.send(fail_silently=True)
        self.assert_esp_not_called("API should not be called when send fails silently")
        self.assertEqual(sent, 0)

    def test_suppress_empty_address_lists(self):
        """Empty cc, bcc, and reply_to shouldn't generate empty headers"""
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("cc", data)
        self.assertNotIn("bcc", data)
        self.assertNotIn("reply_to", data)

    # MailerSend requires at least one "to" address
    # def test_empty_to(self):
    #     self.message.to = []
    #     self.message.cc = ["cc@example.com"]
    #     self.message.send()
    #     data = self.get_api_call_json()

    def test_api_failure(self):
        raw_errors = {
            "message": "Helpful ESP explanation",
            "errors": {"some.field": ["The some.field must be valid."]},
        }
        self.set_mock_response(
            status_code=422, reason="UNPROCESSABLE ENTITY", json_data=raw_errors
        )
        # Error string includes ESP response:
        with self.assertRaisesMessage(AnymailAPIError, "Helpful ESP explanation"):
            self.message.send()

    def test_api_failure_fail_silently(self):
        # Make sure fail_silently is respected
        self.set_mock_response(status_code=422)
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)


@tag("mailersend")
class MailerSendBackendAnymailFeatureTests(MailerSendBackendMockAPITestCase):
    """Test backend support for Anymail added features"""

    def test_envelope_sender(self):
        self.message.envelope_sender = "bounce-handler@bounces.example.com"
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "envelope_sender"):
            self.message.send()

    def test_metadata(self):
        self.message.metadata = {"user_id": "12345", "items": "mailer, send"}
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "metadata"):
            self.message.send()

    def test_send_at(self):
        utc_plus_6 = get_fixed_timezone(6 * 60)
        utc_minus_8 = get_fixed_timezone(-8 * 60)

        with override_current_timezone(utc_plus_6):
            # Timezone-aware datetime converted to UTC:
            self.message.send_at = datetime(2016, 3, 4, 5, 6, 7, tzinfo=utc_minus_8)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["send_at"], timegm((2016, 3, 4, 13, 6, 7))
            )  # 05:06 UTC-8 == 13:06 UTC

            # Timezone-naive datetime assumed to be Django current_timezone
            self.message.send_at = datetime(
                2022, 10, 11, 12, 13, 14, 567
            )  # microseconds should get stripped
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["send_at"], timegm((2022, 10, 11, 6, 13, 14))
            )  # 12:13 UTC+6 == 06:13 UTC

            # Date-only treated as midnight in current timezone
            self.message.send_at = date(2022, 10, 22)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["send_at"], timegm((2022, 10, 21, 18, 0, 0))
            )  # 00:00 UTC+6 == 18:00-1d UTC

            # POSIX timestamp
            self.message.send_at = 1651820889  # 2022-05-06 07:08:09 UTC
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["send_at"], 1651820889)

    def test_tags(self):
        self.message.tags = ["receipt", "repeat-user"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["tags"], ["receipt", "repeat-user"])

    def test_tracking(self):
        # Test one way...
        self.message.track_opens = True
        self.message.track_clicks = False
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["settings"]["track_opens"], True)
        self.assertEqual(data["settings"]["track_clicks"], False)

        # ...and the opposite way
        self.message.track_opens = False
        self.message.track_clicks = True
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["settings"]["track_opens"], False)
        self.assertEqual(data["settings"]["track_clicks"], True)

    def test_template_id(self):
        message = mail.EmailMultiAlternatives(
            from_email="from@example.com", to=["to@example.com"]
        )
        message.template_id = "zyxwvut98765"
        message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["template_id"], "zyxwvut98765")
        # With a template, MailerSend always ignores "text" and "html" params.
        # For "subject", "from_email", and "reply_to" params, MailerSend ignores them
        # if the corresponding field is set in the template's "Default settings",
        # otherwise it uses the params. (And "subject" and "from_email" are required
        # if no default setting for the template.)
        # For "tags", MailerSend uses the param value if provided, otherwise
        # the template default if any. (It does not attempt to combine them.)

    @override_settings(ANYMAIL_MAILERSEND_BATCH_SEND_MODE="expose-to-list")
    def test_merge_data_expose_to_list(self):
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.cc = ["cc@example.com"]
        self.message.body = "Hi {{ name }}. Welcome to {{ group }} at {{ site }}."
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},  # and leave group undefined
            "nobody@example.com": {"name": "Not a recipient for this message"},
        }
        self.message.merge_global_data = {"group": "Users", "site": "ExampleCo"}

        self.message.send()
        # BATCH_SEND_MODE="expose-to-list" uses 'email' API endpoint:
        self.assert_esp_called("/v1/email")
        data = self.get_api_call_json()
        # personalization param covers all recipients:
        self.assertEqual(
            data["personalization"],
            [
                {
                    "email": "alice@example.com",
                    "data": {
                        "name": "Alice",
                        "group": "Developers",
                        "site": "ExampleCo",
                    },
                },
                {
                    "email": "bob@example.com",
                    "data": {"name": "Bob", "group": "Users", "site": "ExampleCo"},
                },
                # No personalization record for "nobody@example.com" -- including a
                # personalization for email not found in "to" param causes API error.
            ],
        )

    @override_settings(ANYMAIL_MAILERSEND_BATCH_SEND_MODE="use-bulk-email")
    def test_merge_data_use_bulk_email(self):
        self.set_mock_response(
            json_data={
                "message": "The bulk email is being processed.",
                "bulk_email_id": "12345abcde",
            }
        )
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.cc = ["cc@example.com"]
        self.message.body = "Hi {{ name }}. Welcome to {{ group }} at {{ site }}."
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},  # and leave group undefined
            "nobody@example.com": {"name": "Not a recipient for this message"},
        }
        self.message.merge_global_data = {"group": "Users", "site": "ExampleCo"}

        self.message.send()
        # BATCH_SEND_MODE="use-bulk-email" uses 'bulk-email' API endpoint:
        self.assert_esp_called("/v1/bulk-email")
        data = self.get_api_call_json()
        self.assertEqual(len(data), 2)  # batch of 2 separate emails
        # "to" split to separate messages:
        self.assertEqual(data[0]["to"], [{"email": "alice@example.com"}])
        self.assertEqual(data[1]["to"], [{"email": "bob@example.com", "name": "Bob"}])
        # "cc" appears in both:
        self.assertEqual(data[0]["cc"], [{"email": "cc@example.com"}])
        self.assertEqual(data[1]["cc"], [{"email": "cc@example.com"}])
        # "personalization" param matches only single recipient:
        self.assertEqual(
            data[0]["personalization"],
            [
                {
                    "email": "alice@example.com",
                    "data": {
                        "name": "Alice",
                        "group": "Developers",
                        "site": "ExampleCo",
                    },
                }
            ],
        )
        self.assertEqual(
            data[1]["personalization"],
            [
                {
                    "email": "bob@example.com",
                    "data": {"name": "Bob", "group": "Users", "site": "ExampleCo"},
                }
            ],
        )

    def test_merge_data_single_recipient(self):
        # BATCH_SEND_MODE=None default uses 'email' for single recipient
        self.message.to = ["Bob <bob@example.com>"]
        self.message.cc = ["cc@example.com"]
        self.message.body = "Hi {{ name }}. Welcome to {{ group }} at {{ site }}."
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},  # and leave group undefined
            "nobody@example.com": {"name": "Not a recipient for this message"},
        }
        self.message.merge_global_data = {"group": "Users", "site": "ExampleCo"}

        self.message.send()
        self.assert_esp_called("/v1/email")
        data = self.get_api_call_json()
        self.assertEqual(
            data["personalization"],
            [
                {
                    "email": "bob@example.com",
                    "data": {"name": "Bob", "group": "Users", "site": "ExampleCo"},
                },
                # No personalization record for merge_data emails not in "to" list.
            ],
        )

    def test_merge_data_ambiguous(self):
        # Multiple recipients require a non-default MAILERSEND_BATCH_SEND_MODE
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.cc = ["cc@example.com"]
        self.message.body = "Hi {{ name }}. Welcome to {{ group }} at {{ site }}."
        # (even an empty merge_data dict should trigger Anymail batch send)
        self.message.merge_data = {}

        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "MAILERSEND_BATCH_SEND_MODE"
        ):
            self.message.send()

    def test_merge_metadata(self):
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123},
            "bob@example.com": {"order_id": 678, "tier": "premium"},
        }

        with self.assertRaisesMessage(AnymailUnsupportedFeature, "merge_metadata"):
            self.message.send()

    def test_default_omits_options(self):
        """Make sure by default we don't send any ESP-specific options.

        Options not specified by the caller should be omitted entirely from
        the API call (*not* sent as False or empty). This ensures
        that your ESP account settings apply by default.
        """
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("cc", data)
        self.assertNotIn("bcc", data)
        self.assertNotIn("reply_to", data)
        self.assertNotIn("html", data)
        self.assertNotIn("attachments", data)
        self.assertNotIn("headers", data)
        self.assertNotIn("template_id", data)
        self.assertNotIn("tags", data)
        self.assertNotIn("variables", data)
        self.assertNotIn("personalization", data)
        self.assertNotIn("precedence_bulk", data)
        self.assertNotIn("send_at", data)
        self.assertNotIn("in_reply_to", data)
        self.assertNotIn("settings", data)

    def test_esp_extra(self):
        self.message.track_clicks = True  # test deep merge of "settings"
        self.message.esp_extra = {
            "variables": [
                {
                    "email": "to@example.com",
                    "substitutions": [{"var": "order_id", "value": "12345"}],
                }
            ],
            "settings": {
                "track_content": True,
            },
        }
        self.message.track_clicks = True
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["variables"],
            [
                {
                    "email": "to@example.com",
                    "substitutions": [{"var": "order_id", "value": "12345"}],
                }
            ],
        )
        self.assertEqual(
            data["settings"],
            {
                "track_content": True,
                "track_clicks": True,  # deep merge
            },
        )

    def test_esp_extra_settings_overrides(self):
        """esp_extra can override batch_send_mode and api_token settings"""
        self.message.merge_data = {}  # trigger batch send
        self.message.esp_extra = {
            "api_token": "token-from-esp-extra",
            "batch_send_mode": "use-bulk-email",
            "hypothetical_future_mailersend_param": 123,
        }
        self.message.send()
        self.assert_esp_called("/v1/bulk-email")  # batch_send_mode from esp_extra
        headers = self.get_api_call_headers()
        self.assertEqual(headers["Authorization"], "Bearer token-from-esp-extra")

        data = self.get_api_call_json()
        self.assertEqual(len(data), 1)  # payload burst for batch
        self.assertNotIn("api_token", data[0])  # not in API payload
        self.assertNotIn("batch_send_mode", data[0])  # not sent to API
        # But other esp_extra params sent:
        self.assertEqual(data[0]["hypothetical_future_mailersend_param"], 123)

    def test_send_attaches_anymail_status(self):
        """The anymail_status should be attached to the message when it is sent"""
        self.set_mock_success(message_id="12345abcde")
        msg = mail.EmailMessage(
            "Subject", "Message", "from@example.com", ["to1@example.com"]
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"queued"})
        self.assertEqual(msg.anymail_status.message_id, "12345abcde")
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "queued"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].message_id,
            "12345abcde",
        )

    @override_settings(
        ANYMAIL_IGNORE_RECIPIENT_STATUS=True  # exception is tested later
    )
    def test_send_all_rejected(self):
        """The anymail_status should be 'rejected' when all recipients rejected"""
        self.set_mock_rejected(
            {"to1@example.com": ["blocklisted"], "to2@example.com": ["hard_bounced"]}
        )
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com", "to2@example.com"],
        )
        msg.send()
        self.assertEqual(msg.anymail_status.status, {"rejected"})
        recipients = msg.anymail_status.recipients
        self.assertEqual(recipients["to1@example.com"].status, "rejected")
        self.assertIsNone(recipients["to1@example.com"].message_id)
        self.assertEqual(recipients["to2@example.com"].status, "rejected")
        self.assertIsNone(recipients["to2@example.com"].message_id)

    def test_send_some_rejected(self):
        """
        The anymail_status should identify which recipients are rejected.
        """
        self.set_mock_rejected(
            {"to1@example.com": ["blocklisted"]},
            warning_type="SOME_SUPPRESSED",
            message_id="12345abcde",
        )
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com", "to2@example.com"],
        )
        msg.send()
        self.assertEqual(msg.anymail_status.status, {"rejected", "queued"})
        recipients = msg.anymail_status.recipients
        self.assertEqual(recipients["to1@example.com"].status, "rejected")
        self.assertIsNone(recipients["to1@example.com"].message_id)
        self.assertEqual(recipients["to2@example.com"].status, "queued")
        self.assertEqual(recipients["to2@example.com"].message_id, "12345abcde")

    # noinspection PyUnresolvedReferences
    @override_settings(ANYMAIL_MAILERSEND_BATCH_SEND_MODE="use-bulk-email")
    def test_bulk_send_response(self):
        self.set_mock_response(
            json_data={
                "message": "The bulk email is being processed.",
                "bulk_email_id": "12345abcde",
            }
        )
        self.message.merge_data = {}  # trigger batch behavior
        self.message.send()
        # Unknown status for bulk send (until you poll the status API):
        self.assertEqual(self.message.anymail_status.status, {"unknown"})
        # Unknown message_id for bulk send, so provide batch id with "bulk:" prefix:
        self.assertEqual(self.message.anymail_status.message_id, "bulk:12345abcde")

    # noinspection PyUnresolvedReferences
    def test_send_failed_anymail_status(self):
        """If the send fails, anymail_status should contain initial values"""
        self.set_mock_response(status_code=400)
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertIsNone(self.message.anymail_status.esp_response)

    # noinspection PyUnresolvedReferences
    def test_unhandled_warnings(self):
        # Non-suppression warnings should turn a 202 accepted response into an error
        response_content = {"warnings": [{"type": "UNKNOWN_WARNING"}]}
        self.set_mock_response(status_code=202, json_data=response_content)
        with self.assertRaisesMessage(AnymailAPIError, "UNKNOWN_WARNING"):
            self.message.send()
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertEqual(
            self.message.anymail_status.esp_response.json(), response_content
        )

    def test_json_serialization_errors(self):
        """Try to provide more information about non-json-serializable data"""
        self.message.tags = [Decimal("19.99")]  # yeah, don't do this
        with self.assertRaises(AnymailSerializationError) as cm:
            self.message.send()
            print(self.get_api_call_json())
        err = cm.exception
        self.assertIsInstance(err, TypeError)  # compatibility with json.dumps
        # our added context:
        self.assertIn("Don't know how to send this data to MailerSend", str(err))
        # original message:
        self.assertRegex(str(err), r"Decimal.*is not JSON serializable")


@tag("mailersend")
class MailerSendBackendRecipientsRefusedTests(MailerSendBackendMockAPITestCase):
    """
    Should raise AnymailRecipientsRefused when *all* recipients are rejected or invalid
    """

    def test_recipients_refused(self):
        self.set_mock_rejected(
            {
                "invalid@localhost": ["hard_bounced"],
                "reject@example.com": ["blocklisted"],
            }
        )
        msg = mail.EmailMessage(
            "Subject",
            "Body",
            "from@example.com",
            ["invalid@localhost", "reject@example.com"],
        )
        with self.assertRaises(AnymailRecipientsRefused):
            msg.send()

    def test_fail_silently(self):
        self.set_mock_rejected(
            {
                "invalid@localhost": ["hard_bounced"],
                "reject@example.com": ["blocklisted"],
            }
        )
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["invalid@localhost", "reject@example.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)

    def test_mixed_response(self):
        """If *any* recipients are valid or queued, no exception is raised"""
        self.set_mock_rejected(
            {
                "invalid@localhost": ["hard_bounced"],
                "reject@example.com": ["blocklisted"],
            },
            warning_type="SOME_SUPPRESSED",
        )
        msg = mail.EmailMessage(
            "Subject",
            "Body",
            "from@example.com",
            [
                "invalid@localhost",
                "valid@example.com",
                "reject@example.com",
                "also.valid@example.com",
            ],
        )
        sent = msg.send()
        # one message sent, successfully, to 2 of 4 recipients:
        self.assertEqual(sent, 1)
        status = msg.anymail_status
        self.assertEqual(status.recipients["invalid@localhost"].status, "rejected")
        self.assertEqual(status.recipients["valid@example.com"].status, "queued")
        self.assertEqual(status.recipients["reject@example.com"].status, "rejected")
        self.assertEqual(status.recipients["also.valid@example.com"].status, "queued")

    @override_settings(ANYMAIL_IGNORE_RECIPIENT_STATUS=True)
    def test_settings_override(self):
        """No exception with ignore setting"""
        self.set_mock_rejected(
            {
                "invalid@localhost": ["hard_bounced"],
                "reject@example.com": ["blocklisted"],
            },
        )
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["invalid@localhost", "reject@example.com"],
        )
        self.assertEqual(sent, 1)  # refused message is included in sent count


@tag("mailersend")
class MailerSendBackendConfigurationTests(MailerSendBackendMockAPITestCase):
    """Test various MailerSend client options"""

    @override_settings(
        # clear MAILERSEND_API_TOKEN from MailerSendBackendMockAPITestCase:
        ANYMAIL={}
    )
    def test_missing_api_token(self):
        with self.assertRaises(AnymailConfigurationError) as cm:
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])
        errmsg = str(cm.exception)
        # Make sure the error mentions the different places to set the key
        self.assertRegex(errmsg, r"\bMAILERSEND_API_TOKEN\b")
        self.assertRegex(errmsg, r"\bANYMAIL_MAILERSEND_API_TOKEN\b")

    @override_settings(
        ANYMAIL={
            "MAILERSEND_API_URL": "https://api.dev.mailersend.com/v2",
            "MAILERSEND_API_TOKEN": "test_api_key",
        }
    )
    def test_mailersend_api_url(self):
        mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])
        self.assert_esp_called("https://api.dev.mailersend.com/v2/email")

        # can also override on individual connection
        connection = mail.get_connection(api_url="https://api.mailersend.com/vNext")
        mail.send_mail(
            "Subject",
            "Message",
            "from@example.com",
            ["to@example.com"],
            connection=connection,
        )
        self.assert_esp_called("https://api.mailersend.com/vNext/email")

    @override_settings(ANYMAIL={"MAILERSEND_API_TOKEN": "bad_token"})
    def test_invalid_api_key(self):
        self.set_mock_response(
            status_code=401,
            reason="UNAUTHORIZED",
            json_data={"message": "Unauthenticated."},
        )
        with self.assertRaisesMessage(AnymailAPIError, "Unauthenticated"):
            self.message.send()



import mimetypes

from ..exceptions import AnymailRequestsAPIError, AnymailUnsupportedFeature
from ..message import AnymailRecipientStatus
from ..utils import CaseInsensitiveCasePreservingDict, get_anymail_setting, update_deep
from .base_requests import AnymailRequestsBackend, RequestsPayload


class EmailBackend(AnymailRequestsBackend):
    """
    MailerSend Email Backend
    """

    esp_name = "MailerSend"

    def __init__(self, **kwargs):
        """Init options from Django settings"""
        esp_name = self.esp_name
        self.api_token = get_anymail_setting(
            "api_token", esp_name=esp_name, kwargs=kwargs, allow_bare=True
        )
        api_url = get_anymail_setting(
            "api_url",
            esp_name=esp_name,
            kwargs=kwargs,
            default="https://api.mailersend.com/v1/",
        )
        if not api_url.endswith("/"):
            api_url += "/"

        #: Can set to "use-bulk-email" or "expose-to-list" or default None
        self.batch_send_mode = get_anymail_setting(
            "batch_send_mode", default=None, esp_name=esp_name, kwargs=kwargs
        )
        super().__init__(api_url, **kwargs)

    def build_message_payload(self, message, defaults):
        return MailerSendPayload(message, defaults, self)

    def parse_recipient_status(self, response, payload, message):
        # The "email" API endpoint responds with an empty text/html body
        # if no warnings, otherwise json with suppression info.
        # The "bulk-email" API endpoint always returns json.
        if response.headers["Content-Type"] == "application/json":
            parsed_response = self.deserialize_json_response(response, payload, message)
        else:
            parsed_response = {}

        try:
            # "email" API endpoint success or SOME_SUPPRESSED
            message_id = response.headers["X-Message-Id"]
            default_status = "queued"
        except KeyError:
            try:
                # "bulk-email" API endpoint
                bulk_id = parsed_response["bulk_email_id"]
                # Add "bulk:" prefix to distinguish from actual message_id.
                message_id = f"bulk:{bulk_id}"
                # Status is determined later; must query API to find out
                default_status = "unknown"
            except KeyError:
                # "email" API endpoint with ALL_SUPPRESSED
                message_id = None
                default_status = "failed"

        # Don't swallow errors (which should have been handled with a non-2xx
        # status, earlier) or any warnings that we won't consume below.
        errors = parsed_response.get("errors", [])
        warnings = parsed_response.get("warnings", [])
        if errors or any(
            warning["type"] not in ("ALL_SUPPRESSED", "SOME_SUPPRESSED")
            for warning in warnings
        ):
            raise AnymailRequestsAPIError(
                "Unexpected MailerSend API response errors/warnings",
                email_message=message,
                payload=payload,
                response=response,
                backend=self,
            )

        # Collect a list of all problem recipients from any suppression warnings.
        # (warnings[].recipients[].reason[] will contain some combination of
        # "hard_bounced", "spam_complaint", "unsubscribed", and/or
        # "blocklisted", all of which map to Anymail's "rejected" status.)
        try:
            # warning["type"] is guaranteed to be {ALL,SOME}_SUPPRESSED at this point.
            rejected_emails = [
                recipient["email"]
                for warning in warnings
                for recipient in warning["recipients"]
            ]
        except (KeyError, TypeError) as err:
            raise AnymailRequestsAPIError(
                f"Unexpected MailerSend API response format: {err!s}",
                email_message=message,
                payload=payload,
                response=response,
                backend=self,
            ) from None

        recipient_status = CaseInsensitiveCasePreservingDict(
            {
                recipient.addr_spec: AnymailRecipientStatus(
                    message_id=message_id, status=default_status
                )
                for recipient in payload.all_recipients
            }
        )
        for rejected_email in rejected_emails:
            recipient_status[rejected_email] = AnymailRecipientStatus(
                message_id=None, status="rejected"
            )

        return dict(recipient_status)


class MailerSendPayload(RequestsPayload):
    def __init__(self, message, defaults, backend, *args, **kwargs):
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            # Token may be changed in set_esp_extra below:
            "Authorization": f"Bearer {backend.api_token}",
        }
        self.all_recipients = []  # needed for parse_recipient_status
        self.merge_data = {}  # late bound
        self.merge_global_data = None  # late bound
        self.batch_send_mode = backend.batch_send_mode  # can override in esp_extra
        super().__init__(message, defaults, backend, headers=headers, *args, **kwargs)

    def get_api_endpoint(self):
        if self.is_batch():
            # MailerSend's "email" endpoint supports per-recipient customizations
            # (merge_data) for batch sending, but exposes the complete "To" list
            # to all recipients. This conflicts with Anymail's batch send model, which
            # expects each recipient can only see their own "To" email.
            #
            # MailerSend's "bulk-email" endpoint can send separate messages to each
            # "To" email, but doesn't return a message_id. (It returns a batch_email_id
            # that can later resolve to message_ids by polling a status API.)
            #
            # Since either of these would cause unexpected behavior, require the user
            # to opt into one via batch_send_mode.
            if self.batch_send_mode == "use-bulk-email":
                return "bulk-email"
            elif self.batch_send_mode == "expose-to-list":
                return "email"
            elif len(self.data["to"]) <= 1:
                # With only one "to", exposing the recipient list is moot.
                # (This covers the common case of single-recipient template merge.)
                return "email"
            else:
                # Unconditionally raise, even if IGNORE_UNSUPPORTED_FEATURES enabled.
                # We can't guess which API to use for this send.
                raise AnymailUnsupportedFeature(
                    f"{self.esp_name} requires MAILERSEND_BATCH_SEND_MODE set to either"
                    f" 'use-bulk-email' or 'expose-to-list' for using batch send"
                    f" (merge_data) with multiple recipients. See the Anymail docs."
                )
        else:
            return "email"

    def serialize_data(self):
        api_endpoint = self.get_api_endpoint()
        needs_personalization = self.merge_data or self.merge_global_data
        if api_endpoint == "email":
            if needs_personalization:
                self.data["personalization"] = [
                    self.personalization_for_email(to["email"])
                    for to in self.data["to"]
                ]
            data = self.data
        elif api_endpoint == "bulk-email":
            # Burst the payload into individual bulk-email recipients:
            data = []
            for to in self.data["to"]:
                recipient_data = self.data.copy()
                recipient_data["to"] = [to]
                if needs_personalization:
                    recipient_data["personalization"] = [
                        self.personalization_for_email(to["email"])
                    ]
                data.append(recipient_data)
        else:
            raise AssertionError(
                f"MailerSendPayload.serialize_data missing"
                f" case for api_endpoint {api_endpoint!r}"
            )
        return self.serialize_json(data)

    def personalization_for_email(self, email):
        """
        Return a MailerSend personalization object for email address.

        Composes merge_global_data and merge_data[email].
        """
        if email in self.merge_data:
            if self.merge_global_data:
                recipient_data = self.merge_global_data.copy()
                recipient_data.update(self.merge_data[email])
            else:
                recipient_data = self.merge_data[email]
        elif self.merge_global_data:
            recipient_data = self.merge_global_data
        else:
            recipient_data = {}
        return {"email": email, "data": recipient_data}

    #
    # Payload construction
    #

    def make_mailersend_email(self, email):
        """Return MailerSend email/name object for an EmailAddress"""
        obj = {"email": email.addr_spec}
        if email.display_name:
            obj["name"] = email.display_name
        return obj

    def init_payload(self):
        self.data = {}  # becomes json

    def set_from_email(self, email):
        self.data["from"] = self.make_mailersend_email(email)

    def set_recipients(self, recipient_type, emails):
        assert recipient_type in ["to", "cc", "bcc"]
        if emails:
            self.data[recipient_type] = [
                self.make_mailersend_email(email) for email in emails
            ]
            self.all_recipients += emails

    def set_subject(self, subject):
        self.data["subject"] = subject

    def set_reply_to(self, emails):
        if len(emails) > 1:
            self.unsupported_feature("multiple reply_to emails")
        elif emails:
            self.data["reply_to"] = self.make_mailersend_email(emails[0])

    def set_extra_headers(self, headers):
        # MailerSend has individual API params for In-Reply-To and Precedence: bulk.
        # The general "headers" option "is available to Enterprise accounts only".
        # (headers is a CaseInsensitiveDict, and is a copy so safe to modify.)
        in_reply_to = headers.pop("In-Reply-To", None)
        if in_reply_to is not None:
            self.data["in_reply_to"] = in_reply_to

        precedence = headers.pop("Precedence", None)
        if precedence is not None:
            # Overrides MailerSend domain-level setting
            is_bulk = precedence.lower() in ("bulk", "junk", "list")
            self.data["precedence_bulk"] = is_bulk

        if headers:
            self.data["headers"] = [
                {"name": field, "value": value} for field, value in headers.items()
            ]

    def set_text_body(self, body):
        self.data["text"] = body

    def set_html_body(self, body):
        if "html" in self.data:
            # second html body could show up through multiple alternatives,
            # or html body + alternative
            self.unsupported_feature("multiple html parts")
        self.data["html"] = body

    def add_attachment(self, attachment):
        # Add a MailerSend attachments[] object for attachment:
        attachment_object = {
            "filename": attachment.name,
            "content": attachment.b64content,
            "disposition": "attachment",
        }
        if not attachment_object["filename"]:
            # MailerSend requires filename, and determines mimetype from it
            # (even for inline attachments). For unnamed attachments, try
            # to generate a generic filename with the correct extension:
            ext = mimetypes.guess_extension(attachment.mimetype, strict=False)
            if ext is not None:
                attachment_object["filename"] = f"attachment{ext}"
        if attachment.inline:
            attachment_object["disposition"] = "inline"
            attachment_object["id"] = attachment.cid
        self.data.setdefault("attachments", []).append(attachment_object)

    # MailerSend doesn't have metadata
    # def set_metadata(self, metadata):

    def set_send_at(self, send_at):
        # Backend has converted pretty much everything to
        # a datetime by here; MailerSend expects unix timestamp
        self.data["send_at"] = int(send_at.timestamp())  # strip microseconds

    def set_tags(self, tags):
        if tags:
            self.data["tags"] = tags

    def set_track_clicks(self, track_clicks):
        self.data.setdefault("settings", {})["track_clicks"] = track_clicks

    def set_track_opens(self, track_opens):
        self.data.setdefault("settings", {})["track_opens"] = track_opens

    def set_template_id(self, template_id):
        self.data["template_id"] = template_id

    def set_merge_data(self, merge_data):
        # late bound in serialize_data
        self.merge_data = merge_data

    def set_merge_global_data(self, merge_global_data):
        # late bound in serialize_data
        self.merge_global_data = merge_global_data

    # MailerSend doesn't have metadata
    # def set_merge_metadata(self, merge_metadata):

    def set_esp_extra(self, extra):
        # Deep merge to allow (e.g.,) {"settings": {"track_content": True}}:
        update_deep(self.data, extra)

        # Allow overriding api_token on individual message:
        try:
            api_token = self.data.pop("api_token")
        except KeyError:
            pass
        else:
            self.headers["Authorization"] = f"Bearer {api_token}"

        # Allow overriding batch_send_mode on individual message:
        try:
            self.batch_send_mode = self.data.pop("batch_send_mode")
        except KeyError:
            pass



import hashlib
import hmac
import json
from datetime import datetime, timezone
from unittest.mock import ANY

from django.core.exceptions import ImproperlyConfigured
from django.test import override_settings, tag

from anymail.exceptions import AnymailConfigurationError
from anymail.signals import AnymailTrackingEvent
from anymail.webhooks.mailgun import MailgunTrackingWebhookView

from .webhook_cases import WebhookBasicAuthTestCase, WebhookTestCase

TEST_WEBHOOK_SIGNING_KEY = "TEST_WEBHOOK_SIGNING_KEY"


def mailgun_signature(timestamp, token, webhook_signing_key):
    """Generates a Mailgun webhook signature"""
    # https://documentation.mailgun.com/en/latest/user_manual.html#securing-webhooks
    return hmac.new(
        key=webhook_signing_key.encode("ascii"),
        msg="{timestamp}{token}".format(timestamp=timestamp, token=token).encode(
            "ascii"
        ),
        digestmod=hashlib.sha256,
    ).hexdigest()


def mailgun_sign_payload(data, webhook_signing_key=TEST_WEBHOOK_SIGNING_KEY):
    """Add or complete Mailgun webhook signature block in data dict"""
    # Modifies the dict in place
    event_data = data.get("event-data", {})
    signature = data.setdefault("signature", {})
    token = signature.setdefault("token", "1234567890abcdef1234567890abcdef")
    timestamp = signature.setdefault(
        "timestamp", str(int(float(event_data.get("timestamp", "1234567890.123"))))
    )
    signature["signature"] = mailgun_signature(
        timestamp, token, webhook_signing_key=webhook_signing_key
    )
    return data


def mailgun_sign_legacy_payload(data, webhook_signing_key=TEST_WEBHOOK_SIGNING_KEY):
    """Add a Mailgun webhook signature to data dict"""
    # Modifies the dict in place
    data.setdefault("timestamp", "1234567890")
    data.setdefault("token", "1234567890abcdef1234567890abcdef")
    data["signature"] = mailgun_signature(
        data["timestamp"], data["token"], webhook_signing_key=webhook_signing_key
    )
    return data


def querydict_to_postdict(qd):
    """Converts a Django QueryDict to a TestClient.post(data)-style dict

    Single-value fields appear as normal
    Multi-value fields appear as a list (differs from QueryDict.dict)
    """
    return {key: values if len(values) > 1 else values[0] for key, values in qd.lists()}


@tag("mailgun")
class MailgunWebhookSettingsTestCase(WebhookTestCase):
    def test_requires_webhook_signing_key(self):
        with self.assertRaisesMessage(
            ImproperlyConfigured, "MAILGUN_WEBHOOK_SIGNING_KEY"
        ):
            self.client.post(
                "/anymail/mailgun/tracking/",
                content_type="application/json",
                data=json.dumps(
                    mailgun_sign_payload({"event-data": {"event": "delivered"}})
                ),
            )

    @override_settings(
        ANYMAIL_MAILGUN_API_KEY="TEST_API_KEY",
        ANYMAIL_MAILGUN_WEBHOOK_SIGNING_KEY="TEST_WEBHOOK_SIGNING_KEY",
    )
    def test_webhook_signing_is_different_from_api_key(self):
        """
        Webhooks should use MAILGUN_WEBHOOK_SIGNING_KEY,
        not MAILGUN_API_KEY, if both provided
        """
        payload = json.dumps(
            mailgun_sign_payload(
                {"event-data": {"event": "delivered"}},
                webhook_signing_key="TEST_WEBHOOK_SIGNING_KEY",
            )
        )
        response = self.client.post(
            "/anymail/mailgun/tracking/", content_type="application/json", data=payload
        )
        self.assertEqual(response.status_code, 200)

    @override_settings(ANYMAIL_MAILGUN_API_KEY="TEST_API_KEY")
    def test_defaults_webhook_signing_to_api_key(self):
        """
        Webhooks should default to MAILGUN_API_KEY
        if MAILGUN_WEBHOOK_SIGNING_KEY not provided
        """
        payload = json.dumps(
            mailgun_sign_payload(
                {"event-data": {"event": "delivered"}},
                webhook_signing_key="TEST_API_KEY",
            )
        )
        response = self.client.post(
            "/anymail/mailgun/tracking/", content_type="application/json", data=payload
        )
        self.assertEqual(response.status_code, 200)

    def test_webhook_signing_key_view_params(self):
        """Webhook signing key can be provided as a view param"""
        view = MailgunTrackingWebhookView.as_view(
            webhook_signing_key="VIEW_SIGNING_KEY"
        )
        view_instance = view.view_class(**view.view_initkwargs)
        self.assertEqual(view_instance.webhook_signing_key, b"VIEW_SIGNING_KEY")

        # Can also use `api_key` param for backwards compatibility
        # with earlier Anymail versions
        view = MailgunTrackingWebhookView.as_view(api_key="VIEW_API_KEY")
        view_instance = view.view_class(**view.view_initkwargs)
        self.assertEqual(view_instance.webhook_signing_key, b"VIEW_API_KEY")


@tag("mailgun")
@override_settings(ANYMAIL_MAILGUN_WEBHOOK_SIGNING_KEY=TEST_WEBHOOK_SIGNING_KEY)
class MailgunWebhookSecurityTestCase(WebhookBasicAuthTestCase):
    should_warn_if_no_auth = False  # because we check webhook signature

    def call_webhook(self):
        return self.client.post(
            "/anymail/mailgun/tracking/",
            content_type="application/json",
            data=json.dumps(
                mailgun_sign_payload({"event-data": {"event": "delivered"}})
            ),
        )

    # Additional tests are in WebhookBasicAuthTestCase

    def test_verifies_correct_signature(self):
        response = self.client.post(
            "/anymail/mailgun/tracking/",
            content_type="application/json",
            data=json.dumps(
                mailgun_sign_payload({"event-data": {"event": "delivered"}})
            ),
        )
        self.assertEqual(response.status_code, 200)

    def test_verifies_missing_signature(self):
        response = self.client.post(
            "/anymail/mailgun/tracking/",
            content_type="application/json",
            data=json.dumps({"event-data": {"event": "delivered"}}),
        )
        self.assertEqual(response.status_code, 400)

    def test_verifies_bad_signature(self):
        data = mailgun_sign_payload(
            {"event-data": {"event": "delivered"}},
            webhook_signing_key="wrong signing key",
        )
        response = self.client.post(
            "/anymail/mailgun/tracking/",
            content_type="application/json",
            data=json.dumps(data),
        )
        self.assertEqual(response.status_code, 400)


@tag("mailgun")
@override_settings(ANYMAIL_MAILGUN_WEBHOOK_SIGNING_KEY=TEST_WEBHOOK_SIGNING_KEY)
class MailgunTestCase(WebhookTestCase):
    # Tests for Mailgun's new webhooks (announced 2018-06-29)

    def test_delivered_event(self):
        # This is an actual, complete (sanitized) "delivered" event as received from
        # Mailgun. (For brevity, later tests omit several payload fields that aren't
        # used by Anymail.)
        raw_event = mailgun_sign_payload(
            {
                "signature": {
                    "timestamp": "1534108637",
                    "token": "651869375b9df3c98fc15c4889b102119add1235c38fc92824",
                    "signature": "...",
                },
                "event-data": {
                    "tags": [],
                    "timestamp": 1534108637.153125,
                    "storage": {
                        "url": "https://sw.api.mailgun.net/v3/domains/"
                        "example.org/messages/eyJwI...",
                        "key": "eyJwI...",
                    },
                    "recipient-domain": "example.com",
                    "id": "hTWCTD81RtiDN-...",
                    "campaigns": [],
                    "user-variables": {},
                    "flags": {
                        "is-routed": False,
                        "is-authenticated": True,
                        "is-system-test": False,
                        "is-test-mode": False,
                    },
                    "log-level": "info",
                    "envelope": {
                        "sending-ip": "333.123.123.200",
                        "sender": "test@example.org",
                        "transport": "smtp",
                        "targets": "recipient@example.com",
                    },
                    "message": {
                        "headers": {
                            "to": "recipient@example.com",
                            "message-id": "20180812211713.1.DF5966851B4BAA99"
                            "@example.org",
                            "from": "test@example.org",
                            "subject": "Testing",
                        },
                        "attachments": [],
                        "size": 809,
                    },
                    "recipient": "recipient@example.com",
                    "event": "delivered",
                    "delivery-status": {
                        "tls": True,
                        "mx-host": "smtp-in.example.com",
                        "attempt-no": 1,
                        "description": "",
                        "session-seconds": 3.5700838565826416,
                        "utf8": True,
                        "code": 250,
                        "message": "OK",
                        "certificate-verified": True,
                    },
                },
            }
        )
        response = self.client.post(
            "/anymail/mailgun/tracking/",
            data=json.dumps(raw_event),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "delivered")
        self.assertEqual(
            event.timestamp,
            datetime(2018, 8, 12, 21, 17, 17, microsecond=153125, tzinfo=timezone.utc),
        )
        self.assertEqual(
            event.message_id, "<20180812211713.1.DF5966851B4BAA99@example.org>"
        )
        # Note that Anymail uses the "token" as its normalized event_id:
        self.assertEqual(
            event.event_id, "651869375b9df3c98fc15c4889b102119add1235c38fc92824"
        )
        # ... if you want the Mailgun "event id",
        # that's available through the raw esp_event:
        self.assertEqual(event.esp_event["event-data"]["id"], "hTWCTD81RtiDN-...")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(event.tags, [])
        self.assertEqual(event.metadata, {})

    def test_failed_permanent_event(self):
        raw_event = mailgun_sign_payload(
            {
                "event-data": {
                    "event": "failed",
                    "severity": "permanent",
                    "reason": "bounce",
                    "recipient": "invalid@example.com",
                    "timestamp": 1534110422.389832,
                    "log-level": "error",
                    "message": {
                        "headers": {
                            "to": "invalid@example.com",
                            "message-id": "20180812214658.1.0DF563D0B3597700"
                            "@example.org",
                            "from": "Test Sender ",
                        },
                    },
                    "delivery-status": {
                        "tls": True,
                        "mx-host": "aspmx.l.example.org",
                        "attempt-no": 1,
                        "description": "",
                        "session-seconds": 2.952177047729492,
                        "utf8": True,
                        "code": 550,
                        "message": "5.1.1 The email account that you tried to reach"
                        " does not exist. Please try\n"
                        "5.1.1 double-checking the recipient's email address for typos",
                        "certificate-verified": True,
                    },
                },
            }
        )
        response = self.client.post(
            "/anymail/mailgun/tracking/",
            data=json.dumps(raw_event),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.recipient, "invalid@example.com")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertEqual(event.description, "")
        self.assertEqual(
            event.mta_response,
            "5.1.1 The email account that you tried to reach does not exist."
            " Please try\n"
            "5.1.1 double-checking the recipient's email address for typos",
        )

    def test_failed_temporary_event(self):
        raw_event = mailgun_sign_payload(
            {
                "event-data": {
                    "event": "failed",
                    "severity": "temporary",
                    "reason": "generic",
                    "timestamp": 1534111899.659519,
                    "log-level": "warn",
                    "message": {
                        "headers": {
                            "to": "undeliverable@nomx.example.com",
                            "message-id": "20180812214638.1.4A7D468E9BC18C5D"
                            "@example.org",
                            "from": "Test Sender ",
                            "subject": "Testing",
                        },
                    },
                    "recipient": "undeliverable@nomx.example.com",
                    "delivery-status": {
                        "attempt-no": 3,
                        "description": "No MX for nomx.example.com",
                        "session-seconds": 0.0,
                        "retry-seconds": 1800,
                        "code": 498,
                        "message": "No MX for nomx.example.com",
                    },
                },
            }
        )
        response = self.client.post(
            "/anymail/mailgun/tracking/",
            data=json.dumps(raw_event),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "deferred")
        self.assertEqual(event.recipient, "undeliverable@nomx.example.com")
        self.assertEqual(event.reject_reason, "other")
        self.assertEqual(event.description, "No MX for nomx.example.com")
        self.assertEqual(event.mta_response, "No MX for nomx.example.com")

    def test_failed_greylisted_event(self):
        raw_event = mailgun_sign_payload(
            {
                "event-data": {
                    "event": "failed",
                    "severity": "temporary",
                    "reason": "greylisted",
                    "timestamp": 1534111899.659519,
                    "log-level": "warn",
                    "message": {
                        "headers": {
                            "to": "undeliverable@nomx.example.com",
                            "message-id": "20180812214638.1.4A7D468E9BC18C5D"
                            "@example.org",
                            "from": "Test Sender ",
                            "subject": "Testing",
                        },
                    },
                    "recipient": "undeliverable@mx.example.com",
                    "delivery-status": {
                        "mx-host": "mx.example.com",
                        "attempt-no": 1,
                        "description": "Recipient address rejected: Greylisted",
                        "session-seconds": 0.0,
                        "retry-seconds": 300,
                        "code": 450,
                        "message": "Recipient address rejected: Greylisted",
                    },
                },
            }
        )
        response = self.client.post(
            "/anymail/mailgun/tracking/",
            data=json.dumps(raw_event),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "deferred")
        self.assertEqual(event.recipient, "undeliverable@mx.example.com")
        self.assertEqual(event.reject_reason, "other")
        self.assertEqual(event.description, "Recipient address rejected: Greylisted")
        self.assertEqual(event.mta_response, "Recipient address rejected: Greylisted")

    def test_rejected_event(self):
        # (The "rejected" event is documented and appears in Mailgun dashboard logs,
        # but it doesn't appear to be delivered through webhooks as of 8/2018.)
        # Note that this payload lacks the recipient field present in all other events.
        raw_event = mailgun_sign_payload(
            {
                "event-data": {
                    "event": "rejected",
                    "timestamp": 1529704976.104692,
                    "log-level": "warn",
                    "reject": {
                        "reason": "Sandbox subdomains are for test purposes only.",
                        "description": "",
                    },
                    "message": {
                        "headers": {
                            "to": "Recipient Name <recipient@example.org>",
                            "message-id": "20180622220256.1.B31A451A2E5422BB"
                            "@sandbox55887.mailgun.org",
                            "from": "test@sandbox55887.mailgun.org",
                            "subject": "Test Subject",
                        },
                    },
                },
            }
        )
        response = self.client.post(
            "/anymail/mailgun/tracking/",
            data=json.dumps(raw_event),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "rejected")
        self.assertEqual(event.reject_reason, "other")
        self.assertEqual(
            event.description, "Sandbox subdomains are for test purposes only."
        )
        self.assertEqual(event.recipient, "recipient@example.org")

    def test_complained_event(self):
        raw_event = mailgun_sign_payload(
            {
                "event-data": {
                    "event": "complained",
                    "id": "ncV2XwymRUKbPek_MIM-Gw",
                    "timestamp": 1377214260.049634,
                    "log-level": "warn",
                    "recipient": "recipient@example.com",
                    "message": {
                        "headers": {
                            "to": "foo@recipient.com",
                            "message-id": "20130718032413.263EE2E0926@example.org",
                            "from": "Sender Name <sender@example.org>",
                            "subject": "We are not spammer",
                        },
                    },
                },
            }
        )
        response = self.client.post(
            "/anymail/mailgun/tracking/",
            data=json.dumps(raw_event),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "complained")
        self.assertEqual(event.recipient, "recipient@example.com")

    def test_unsubscribed_event(self):
        raw_event = mailgun_sign_payload(
            {
                "event-data": {
                    "event": "unsubscribed",
                    "id": "W3X4JOhFT-OZidZGKKr9iA",
                    "timestamp": 1377213791.421473,
                    "log-level": "info",
                    "recipient": "recipient@example.com",
                    "message": {
                        "headers": {
                            "message-id": "20130822232216.13966.79700"
                            "@samples.mailgun.org"
                        }
                    },
                },
            }
        )
        response = self.client.post(
            "/anymail/mailgun/tracking/",
            data=json.dumps(raw_event),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "unsubscribed")
        self.assertEqual(event.recipient, "recipient@example.com")

    def test_opened_event(self):
        raw_event = mailgun_sign_payload(
            {
                "event-data": {
                    "event": "opened",
                    "timestamp": 1534109600.089676,
                    "recipient": "recipient@example.com",
                    "tags": ["welcome", "variation-A"],
                    "user-variables": {"cohort": "2018-08-B", "user_id": "123456"},
                    "message": {
                        # Mailgun *only* includes the message-id header
                        # for opened, clicked events...
                        "headers": {
                            "message-id": "20180812213139.1"
                            ".BC6694A917BB7E6A@example.org"
                        }
                    },
                    "geolocation": {
                        "country": "US",
                        "region": "CA",
                        "city": "San Francisco",
                    },
                    "ip": "888.222.444.111",
                    "client-info": {
                        "client-type": "browser",
                        "client-os": "OS X",
                        "device-type": "desktop",
                        "client-name": "Chrome",
                        "user-agent": "Mozilla/5.0"
                        " (Macintosh; Intel Mac OS X 10_13_6)...",
                    },
                }
            }
        )
        response = self.client.post(
            "/anymail/mailgun/tracking/",
            data=json.dumps(raw_event),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "opened")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(event.tags, ["welcome", "variation-A"])
        self.assertEqual(event.metadata, {"cohort": "2018-08-B", "user_id": "123456"})

    def test_clicked_event(self):
        raw_event = mailgun_sign_payload(
            {
                "event-data": {
                    "event": "clicked",
                    "timestamp": 1534109600.089676,
                    "recipient": "recipient@example.com",
                    "url": "https://example.com/test",
                }
            }
        )
        response = self.client.post(
            "/anymail/mailgun/tracking/",
            data=json.dumps(raw_event),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "clicked")
        self.assertEqual(event.click_url, "https://example.com/test")

    def test_delivery_status_is_none_event(self):
        raw_event = mailgun_sign_payload(
            {
                "event-data": {
                    "event": "accepted",
                    "delivery-status": None,
                }
            }
        )
        response = self.client.post(
            "/anymail/mailgun/tracking/",
            data=json.dumps(raw_event),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)


@tag("mailgun")
@override_settings(ANYMAIL_MAILGUN_WEBHOOK_SIGNING_KEY=TEST_WEBHOOK_SIGNING_KEY)
class MailgunLegacyTestCase(WebhookTestCase):
    # Tests for Mailgun's "legacy" webhooks
    # (which were the only webhooks available prior to Anymail 4.0)

    def test_delivered_event(self):
        raw_event = mailgun_sign_legacy_payload(
            {
                "domain": "example.com",
                "message-headers": json.dumps(
                    [
                        ["Sender", "from=example.com"],
                        ["Date", "Thu, 21 Apr 2016 17:55:29 +0000"],
                        [
                            "X-Mailgun-Sid",
                            "WyIxZmY4ZSIsICJtZWRtdW5kc0BnbWFpbC5jb20iLCAiZjFjNzgyIl0=",
                        ],
                        [
                            "Received",
                            "by luna.mailgun.net with HTTP;"
                            " Thu, 21 Apr 2016 17:55:29 +0000",
                        ],
                        [
                            "Message-Id",
                            "<20160421175529.19495.89030.B3AE3728@example.com>",
                        ],
                        ["To", "recipient@example.com"],
                        ["From", "from@example.com"],
                        ["Subject", "Webhook testing"],
                        ["Mime-Version", "1.0"],
                        [
                            "Content-Type",
                            [
                                "multipart/alternative",
                                {"boundary": "74fb561763da440d8e6a034054974251"},
                            ],
                        ],
                    ]
                ),
                "X-Mailgun-Sid": "WyIxZmY4ZSIsICJtZWRtdW5kc"
                "0BnbWFpbC5jb20iLCAiZjFjNzgyIl0=",
                "token": "06c96bafc3f42a66b9edd546347a2fe18dc23461fe80dc52f0",
                "timestamp": "1461261330",
                "Message-Id": "<20160421175529.19495.89030.B3AE3728@example.com>",
                "recipient": "recipient@example.com",
                "event": "delivered",
            }
        )
        response = self.client.post("/anymail/mailgun/tracking/", data=raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "delivered")
        self.assertEqual(
            event.timestamp, datetime(2016, 4, 21, 17, 55, 30, tzinfo=timezone.utc)
        )
        self.assertEqual(
            event.message_id, "<20160421175529.19495.89030.B3AE3728@example.com>"
        )
        self.assertEqual(
            event.event_id, "06c96bafc3f42a66b9edd546347a2fe18dc23461fe80dc52f0"
        )
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(querydict_to_postdict(event.esp_event), raw_event)
        self.assertEqual(event.tags, [])
        self.assertEqual(event.metadata, {})

    def test_dropped_bounce(self):
        raw_event = mailgun_sign_legacy_payload(
            {
                "code": "605",
                "domain": "example.com",
                "description": "Not delivering to previously bounced address",
                "attachment-count": "1",
                "Message-Id": "<20160421180324.70521.79375.96884DDB@example.com>",
                "reason": "hardfail",
                "event": "dropped",
                "message-headers": json.dumps(
                    [
                        [
                            "X-Mailgun-Sid",
                            "WyI3Y2VjMyIsICJib3VuY2VAZXhhbXBsZS5jb20iLCAiZjFjNzgyIl0=",
                        ],
                        [
                            "Received",
                            "by luna.mailgun.net with HTTP;"
                            " Thu, 21 Apr 2016 18:03:24 +0000",
                        ],
                        [
                            "Message-Id",
                            "<20160421180324.70521.79375.96884DDB@example.com>",
                        ],
                        ["To", "bounce@example.com"],
                        ["From", "from@example.com"],
                        ["Subject", "Webhook testing"],
                        ["Mime-Version", "1.0"],
                        [
                            "Content-Type",
                            [
                                "multipart/alternative",
                                {"boundary": "a5b51388a4e3455d8feb8510bb8c9fa2"},
                            ],
                        ],
                    ]
                ),
                "recipient": "bounce@example.com",
                "timestamp": "1461261330",
                "X-Mailgun-Sid": "WyI3Y2VjMyIsICJib3VuY2VAZ"
                "XhhbXBsZS5jb20iLCAiZjFjNzgyIl0=",
                "token": "a3fe1fa1640349ac552b84ddde373014b4c41645830c8dd3fc",
            }
        )
        response = self.client.post("/anymail/mailgun/tracking/", data=raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "rejected")
        self.assertEqual(
            event.timestamp, datetime(2016, 4, 21, 17, 55, 30, tzinfo=timezone.utc)
        )
        self.assertEqual(
            event.message_id, "<20160421180324.70521.79375.96884DDB@example.com>"
        )
        self.assertEqual(
            event.event_id, "a3fe1fa1640349ac552b84ddde373014b4c41645830c8dd3fc"
        )
        self.assertEqual(event.recipient, "bounce@example.com")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertEqual(
            event.description, "Not delivering to previously bounced address"
        )
        self.assertEqual(querydict_to_postdict(event.esp_event), raw_event)

    def test_dropped_spam(self):
        raw_event = mailgun_sign_legacy_payload(
            {
                "code": "607",
                "description": "Not delivering to a user who marked"
                " your messages as spam",
                "reason": "hardfail",
                "event": "dropped",
                "recipient": "complaint@example.com",
                # (omitting some fields that aren't relevant to the test)
            }
        )
        response = self.client.post("/anymail/mailgun/tracking/", data=raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "rejected")
        self.assertEqual(event.reject_reason, "spam")
        self.assertEqual(
            event.description,
            "Not delivering to a user who marked your messages as spam",
        )

    def test_dropped_timed_out(self):
        raw_event = mailgun_sign_legacy_payload(
            {
                "code": "499",
                "description": "Unable to connect to MX servers: [example.com]",
                "reason": "old",
                "event": "dropped",
                "recipient": "complaint@example.com",
                # (omitting some fields that aren't relevant to the test)
            }
        )
        response = self.client.post("/anymail/mailgun/tracking/", data=raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "rejected")
        self.assertEqual(event.reject_reason, "timed_out")
        self.assertEqual(
            event.description, "Unable to connect to MX servers: [example.com]"
        )

    def test_invalid_mailbox(self):
        raw_event = mailgun_sign_legacy_payload(
            {
                "code": "550",
                "error": "550 5.1.1 The email account that you tried to reach does"
                " not exist. Please try "
                "    5.1.1 double-checking the recipient's email address for typos or "
                "    5.1.1 unnecessary spaces.",
                "event": "bounced",
                "recipient": "noreply@example.com",
                # (omitting some fields that aren't relevant to the test)
            }
        )
        response = self.client.post("/anymail/mailgun/tracking/", data=raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertIn(
            "The email account that you tried to reach does not exist",
            event.mta_response,
        )

    def test_alt_smtp_code(self):
        # In some cases, Mailgun uses RFC-3463 extended SMTP status codes
        # (x.y.z, rather than nnn). See issue #62.
        raw_event = mailgun_sign_legacy_payload(
            {
                "code": "5.1.1",
                "error": "smtp;550 5.1.1 RESOLVER.ADR.RecipNotFound; not found",
                "event": "bounced",
                "recipient": "noreply@example.com",
                # (omitting some fields that aren't relevant to the test)
            }
        )
        response = self.client.post("/anymail/mailgun/tracking/", data=raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailgunTrackingWebhookView,
            event=ANY,
            esp_name="Mailgun",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertIn("RecipNotFound", event.mta_response)

    def test_metadata_message_headers(self):
        # Metadata fields are interspersed with other data, but also in message-headers
        # for delivered, bounced and dropped events
        raw_event = mailgun_sign_legacy_payload(
            {
                "event": "delivered",
                "message-headers": json.dumps(
                    [
                        [
                            "X-Mailgun-Variables",
                            '{"custom1": "value1",'
                            ' "custom2": "{\\"key\\":\\"value\\"}"}',
                        ],
                    ]
                ),
                "custom1": "value1",
                # you can store JSON, but you'll need to unpack it yourself:
                "custom2": '{"key":"value"}',
            }
        )
        self.client.post("/anymail/mailgun/tracking/", data=raw_event)
        kwargs = self.assert_handler_called_once_with(self.tracking_handler)
        event = kwargs["event"]
        self.assertEqual(
            event.metadata, {"custom1": "value1", "custom2": '{"key":"value"}'}
        )

    def test_metadata_post_fields(self):
        # Metadata fields are only interspersed with other event params
        # for opened, clicked, unsubscribed events
        raw_event = mailgun_sign_legacy_payload(
            {
                "event": "clicked",
                "custom1": "value1",
                # you can store JSON, but you'll need to unpack it yourself:
                "custom2": '{"key":"value"}',
            }
        )
        self.client.post("/anymail/mailgun/tracking/", data=raw_event)
        kwargs = self.assert_handler_called_once_with(self.tracking_handler)
        event = kwargs["event"]
        self.assertEqual(
            event.metadata, {"custom1": "value1", "custom2": '{"key":"value"}'}
        )

    def test_metadata_key_conflicts(self):
        # If you happen to name metadata (user-variable) keys the same as Mailgun
        # event properties, Mailgun will include both in the webhook post.
        # Make sure we don't confuse them.
        metadata = {
            "event": "metadata-event",
            "recipient": "metadata-recipient",
            "signature": "metadata-signature",
            "timestamp": "metadata-timestamp",
            "token": "metadata-token",
            "ordinary field": "ordinary metadata value",
        }

        raw_event = mailgun_sign_legacy_payload(
            {
                "event": "clicked",
                "recipient": "actual-recipient@example.com",
                "token": "actual-event-token",
                "timestamp": "1461261330",
                "url": "http://clicked.example.com/actual/event/param",
                "h": "an (undocumented) Mailgun event param",
                "tag": ["actual-tag-1", "actual-tag-2"],
            }
        )

        # Simulate how Mailgun merges user-variables fields into event:
        for key in metadata.keys():
            if key in raw_event:
                if key in {"signature", "timestamp", "token"}:
                    # For these fields, Mailgun's value appears after the metadata value
                    raw_event[key] = [metadata[key], raw_event[key]]
                elif key == "message-headers":
                    pass  # Mailgun won't merge this field into the event
                else:
                    # For all other fields, the defined event value comes first
                    raw_event[key] = [raw_event[key], metadata[key]]
            else:
                raw_event[key] = metadata[key]

        response = self.client.post("/anymail/mailgun/tracking/", data=raw_event)
        # if this fails, signature checking is using metadata values:
        self.assertEqual(response.status_code, 200)

        kwargs = self.assert_handler_called_once_with(self.tracking_handler)
        event = kwargs["event"]
        self.assertEqual(event.event_type, "clicked")
        self.assertEqual(event.recipient, "actual-recipient@example.com")
        self.assertEqual(event.timestamp.isoformat(), "2016-04-21T17:55:30+00:00")
        self.assertEqual(event.event_id, "actual-event-token")
        self.assertEqual(event.tags, ["actual-tag-1", "actual-tag-2"])
        self.assertEqual(event.metadata, metadata)

    def test_tags(self):
        # Most events include multiple 'tag' fields for message's tags
        raw_event = mailgun_sign_legacy_payload(
            {
                # Django TestClient encodes list as multiple field values:
                "tag": ["tag1", "tag2"],
                "event": "opened",
            }
        )
        self.client.post("/anymail/mailgun/tracking/", data=raw_event)
        kwargs = self.assert_handler_called_once_with(self.tracking_handler)
        event = kwargs["event"]
        self.assertEqual(event.tags, ["tag1", "tag2"])

    def test_x_tags(self):
        # Delivery events don't include 'tag', but do include 'X-Mailgun-Tag' fields
        raw_event = mailgun_sign_legacy_payload(
            {
                "X-Mailgun-Tag": ["tag1", "tag2"],
                "event": "delivered",
            }
        )
        self.client.post("/anymail/mailgun/tracking/", data=raw_event)
        kwargs = self.assert_handler_called_once_with(self.tracking_handler)
        event = kwargs["event"]
        self.assertEqual(event.tags, ["tag1", "tag2"])

    def test_misconfigured_inbound(self):
        raw_event = mailgun_sign_legacy_payload(
            {
                "recipient": "test@inbound.example.com",
                "sender": "envelope-from@example.org",
                "message-headers": "[]",
                "body-plain": "Test body plain",
                "body-html": "<div>Test body html</div>",
            }
        )

        with self.assertRaisesMessage(
            AnymailConfigurationError,
            "You seem to have set Mailgun's *inbound* route"
            " to Anymail's Mailgun *tracking* webhook URL.",
        ):
            self.client.post("/anymail/mailgun/tracking/", data=raw_event)



import json
from base64 import b64encode
from datetime import datetime, timezone
from textwrap import dedent
from unittest.mock import ANY, patch

from django.test import tag

from anymail.exceptions import AnymailAPIError, AnymailConfigurationError
from anymail.inbound import AnymailInboundMessage
from anymail.signals import AnymailInboundEvent
from anymail.webhooks.amazon_ses import AmazonSESInboundWebhookView

from .test_amazon_ses_webhooks import AmazonSESWebhookTestsMixin
from .webhook_cases import WebhookTestCase


@tag("amazon_ses")
class AmazonSESInboundTests(WebhookTestCase, AmazonSESWebhookTestsMixin):
    def setUp(self):
        super().setUp()
        # Mock boto3.session.Session().client('s3').download_fileobj. (We could also
        # use botocore.stub.Stubber, but mock works well with our test structure.)
        self.patch_boto3_session = patch(
            "anymail.webhooks.amazon_ses.boto3.session.Session", autospec=True
        )
        self.mock_session = self.patch_boto3_session.start()  # boto3.session.Session
        self.addCleanup(self.patch_boto3_session.stop)

        def mock_download_fileobj(bucket, key, fileobj):
            fileobj.write(self.mock_s3_downloadables[bucket][key])

        self.mock_s3_downloadables = {}  #: bucket: key: bytes
        #: boto3.session.Session().client
        self.mock_client = self.mock_session.return_value.client
        #: boto3.session.Session().client('s3', ...)
        self.mock_s3 = self.mock_client.return_value
        self.mock_s3.download_fileobj.side_effect = mock_download_fileobj

    TEST_MIME_MESSAGE = dedent(
        """\
        Return-Path: <bounce-handler@mail.example.org>
        Received: from mail.example.org by inbound-smtp.us-east-1.amazonaws.com...
        MIME-Version: 1.0
        Received: by 10.1.1.1 with HTTP; Fri, 30 Mar 2018 10:21:49 -0700 (PDT)
        From: "Sender, Inc." <from@example.org>
        Date: Fri, 30 Mar 2018 10:21:50 -0700
        Message-ID: <CAEPk3RKsi@mail.example.org>
        Subject: Test inbound message
        To: Recipient <inbound@example.com>, someone-else@example.org
        Content-Type: multipart/alternative; boundary="94eb2c05e174adb140055b6339c5"

        --94eb2c05e174adb140055b6339c5
        Content-Type: text/plain; charset="UTF-8"
        Content-Transfer-Encoding: quoted-printable

        It's a body=E2=80=A6

        --94eb2c05e174adb140055b6339c5
        Content-Type: text/html; charset="UTF-8"
        Content-Transfer-Encoding: quoted-printable

        <div dir=3D"ltr">It's a body=E2=80=A6</div>

        --94eb2c05e174adb140055b6339c5--
        """
    ).replace("\n", "\r\n")

    def test_inbound_sns_utf8(self):
        raw_ses_event = {
            "notificationType": "Received",
            "mail": {
                "timestamp": "2018-03-30T17:21:51.636Z",
                "source": "envelope-from@example.org",
                # messageId is assigned by Amazon SES:
                "messageId": "jili9m351il3gkburn7o2f0u6788stij94c8ld01",
                "destination": ["inbound@example.com", "someone-else@example.org"],
                "headersTruncated": False,
                "headers": [
                    # (omitting a few headers that Amazon SES adds on receipt)
                    {
                        "name": "Return-Path",
                        "value": "<bounce-handler@mail.example.org>",
                    },
                    {
                        "name": "Received",
                        "value": "from mail.example.org by"
                        " inbound-smtp.us-east-1.amazonaws.com...",
                    },
                    {"name": "MIME-Version", "value": "1.0"},
                    {
                        "name": "Received",
                        "value": "by 10.1.1.1 with HTTP;"
                        " Fri, 30 Mar 2018 10:21:49 -0700 (PDT)",
                    },
                    {"name": "From", "value": '"Sender, Inc." <from@example.org>'},
                    {"name": "Date", "value": "Fri, 30 Mar 2018 10:21:50 -0700"},
                    {"name": "Message-ID", "value": "<CAEPk3RKsi@mail.example.org>"},
                    {"name": "Subject", "value": "Test inbound message"},
                    {
                        "name": "To",
                        "value": "Recipient <inbound@example.com>,"
                        " someone-else@example.org",
                    },
                    {
                        "name": "Content-Type",
                        "value": "multipart/alternative;"
                        ' boundary="94eb2c05e174adb140055b6339c5"',
                    },
                ],
                "commonHeaders": {
                    "returnPath": "bounce-handler@mail.example.org",
                    "from": ['"Sender, Inc." <from@example.org>'],
                    "date": "Fri, 30 Mar 2018 10:21:50 -0700",
                    "to": [
                        "Recipient <inbound@example.com>",
                        "someone-else@example.org",
                    ],
                    "messageId": "<CAEPk3RKsi@mail.example.org>",
                    "subject": "Test inbound message",
                },
            },
            "receipt": {
                "timestamp": "2018-03-30T17:21:51.636Z",
                "processingTimeMillis": 357,
                "recipients": ["inbound@example.com"],
                "spamVerdict": {"status": "PASS"},
                "virusVerdict": {"status": "PASS"},
                "spfVerdict": {"status": "PASS"},
                "dkimVerdict": {"status": "PASS"},
                "dmarcVerdict": {"status": "PASS"},
                "action": {
                    "type": "SNS",
                    "topicArn": "arn:aws:sns:us-east-1:111111111111:SES_Inbound",
                    "encoding": "UTF8",
                },
            },
            "content": self.TEST_MIME_MESSAGE,
        }

        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "8f6dee70-c885-558a-be7d-bd48bbf5335e",
            "TopicArn": "arn:aws:sns:us-east-1:111111111111:SES_Inbound",
            "Subject": "Amazon SES Email Receipt Notification",
            "Message": json.dumps(raw_ses_event),
            "Timestamp": "2018-03-30T17:17:36.516Z",
            "SignatureVersion": "1",
            "Signature": "EXAMPLE_SIGNATURE==",
            "SigningCertURL": "https://sns.us-east-1.amazonaws.com"
            "/SimpleNotificationService-12345abcde.pem",
            "UnsubscribeURL": "https://sns.us-east-1.amazonaws.com"
            "/?Action=Unsubscribe&SubscriptionArn=arn...",
        }

        response = self.post_from_sns("/anymail/amazon_ses/inbound/", raw_sns_message)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=AmazonSESInboundWebhookView,
            event=ANY,
            esp_name="Amazon SES",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailInboundEvent)
        self.assertEqual(event.event_type, "inbound")
        self.assertEqual(
            event.timestamp,
            datetime(2018, 3, 30, 17, 21, 51, microsecond=636000, tzinfo=timezone.utc),
        )
        self.assertEqual(event.event_id, "jili9m351il3gkburn7o2f0u6788stij94c8ld01")
        self.assertIsInstance(event.message, AnymailInboundMessage)
        self.assertEqual(event.esp_event, raw_ses_event)

        message = event.message
        self.assertIsInstance(message, AnymailInboundMessage)
        self.assertEqual(message.envelope_sender, "envelope-from@example.org")
        self.assertEqual(message.envelope_recipient, "inbound@example.com")
        self.assertEqual(str(message.from_email), '"Sender, Inc." <from@example.org>')
        self.assertEqual(
            [str(to) for to in message.to],
            ["Recipient <inbound@example.com>", "someone-else@example.org"],
        )
        self.assertEqual(message.subject, "Test inbound message")
        self.assertEqual(message.text, "It's a body\N{HORIZONTAL ELLIPSIS}\r\n")
        self.assertEqual(
            message.html,
            """<div dir="ltr">It's a body\N{HORIZONTAL ELLIPSIS}</div>\r\n""",
        )
        self.assertIs(message.spam_detected, False)

    def test_inbound_sns_base64(self):
        """Should handle 'Base 64' content option on received email SNS action"""
        raw_ses_event = {
            # (omitting some fields that aren't used by Anymail)
            "notificationType": "Received",
            "mail": {
                "source": "envelope-from@example.org",
                "timestamp": "2018-03-30T17:21:51.636Z",
                # messageId is assigned by Amazon SES
                "messageId": "jili9m351il3gkburn7o2f0u6788stij94c8ld01",
                "destination": ["inbound@example.com", "someone-else@example.org"],
            },
            "receipt": {
                "recipients": ["inbound@example.com"],
                "action": {
                    "type": "SNS",
                    "topicArn": "arn:aws:sns:us-east-1:111111111111:SES_Inbound",
                    "encoding": "BASE64",
                },
                "spamVerdict": {"status": "FAIL"},
            },
            "content": b64encode(self.TEST_MIME_MESSAGE.encode("ascii")).decode(
                "ascii"
            ),
        }

        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "8f6dee70-c885-558a-be7d-bd48bbf5335e",
            "TopicArn": "arn:aws:sns:us-east-1:111111111111:SES_Inbound",
            "Message": json.dumps(raw_ses_event),
        }

        response = self.post_from_sns("/anymail/amazon_ses/inbound/", raw_sns_message)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=AmazonSESInboundWebhookView,
            event=ANY,
            esp_name="Amazon SES",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailInboundEvent)
        self.assertEqual(event.event_type, "inbound")
        self.assertEqual(
            event.timestamp,
            datetime(2018, 3, 30, 17, 21, 51, microsecond=636000, tzinfo=timezone.utc),
        )
        self.assertEqual(event.event_id, "jili9m351il3gkburn7o2f0u6788stij94c8ld01")
        self.assertIsInstance(event.message, AnymailInboundMessage)
        self.assertEqual(event.esp_event, raw_ses_event)

        message = event.message
        self.assertIsInstance(message, AnymailInboundMessage)
        self.assertEqual(message.envelope_sender, "envelope-from@example.org")
        self.assertEqual(message.envelope_recipient, "inbound@example.com")
        self.assertEqual(str(message.from_email), '"Sender, Inc." <from@example.org>')
        self.assertEqual(
            [str(to) for to in message.to],
            ["Recipient <inbound@example.com>", "someone-else@example.org"],
        )
        self.assertEqual(message.subject, "Test inbound message")
        self.assertEqual(message.text, "It's a body\N{HORIZONTAL ELLIPSIS}\r\n")
        self.assertEqual(
            message.html,
            """<div dir="ltr">It's a body\N{HORIZONTAL ELLIPSIS}</div>\r\n""",
        )
        self.assertIs(message.spam_detected, True)

    def test_inbound_s3(self):
        """Should handle 'S3' receipt action"""

        self.mock_s3_downloadables["InboundEmailBucket-KeepPrivate"] = {
            "inbound/fqef5sop459utgdf4o9lqbsv7jeo73pejig34301": (
                self.TEST_MIME_MESSAGE.encode("ascii")
            )
        }

        raw_ses_event = {
            # (omitting some fields that aren't used by Anymail)
            "notificationType": "Received",
            "mail": {
                "source": "envelope-from@example.org",
                "timestamp": "2018-03-30T17:21:51.636Z",
                # messageId is assigned by Amazon SES
                "messageId": "fqef5sop459utgdf4o9lqbsv7jeo73pejig34301",
                "destination": ["inbound@example.com", "someone-else@example.org"],
            },
            "receipt": {
                "recipients": ["inbound@example.com"],
                "action": {
                    "type": "S3",
                    "topicArn": "arn:aws:sns:us-east-1:111111111111:SES_Inbound",
                    "bucketName": "InboundEmailBucket-KeepPrivate",
                    "objectKeyPrefix": "inbound",
                    "objectKey": "inbound/fqef5sop459utgdf4o9lqbsv7jeo73pejig34301",
                },
                "spamVerdict": {"status": "GRAY"},
            },
        }
        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "8f6dee70-c885-558a-be7d-bd48bbf5335e",
            "TopicArn": "arn:aws:sns:us-east-1:111111111111:SES_Inbound",
            "Message": json.dumps(raw_ses_event),
        }
        response = self.post_from_sns("/anymail/amazon_ses/inbound/", raw_sns_message)
        self.assertEqual(response.status_code, 200)

        self.mock_client.assert_called_once_with("s3", config=ANY)
        self.mock_s3.download_fileobj.assert_called_once_with(
            "InboundEmailBucket-KeepPrivate",
            "inbound/fqef5sop459utgdf4o9lqbsv7jeo73pejig34301",
            ANY,
        )

        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=AmazonSESInboundWebhookView,
            event=ANY,
            esp_name="Amazon SES",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailInboundEvent)
        self.assertEqual(event.event_type, "inbound")
        self.assertEqual(
            event.timestamp,
            datetime(2018, 3, 30, 17, 21, 51, microsecond=636000, tzinfo=timezone.utc),
        )
        self.assertEqual(event.event_id, "fqef5sop459utgdf4o9lqbsv7jeo73pejig34301")
        self.assertIsInstance(event.message, AnymailInboundMessage)
        self.assertEqual(event.esp_event, raw_ses_event)

        message = event.message
        self.assertIsInstance(message, AnymailInboundMessage)
        self.assertEqual(message.envelope_sender, "envelope-from@example.org")
        self.assertEqual(message.envelope_recipient, "inbound@example.com")
        self.assertEqual(str(message.from_email), '"Sender, Inc." <from@example.org>')
        self.assertEqual(
            [str(to) for to in message.to],
            ["Recipient <inbound@example.com>", "someone-else@example.org"],
        )
        self.assertEqual(message.subject, "Test inbound message")
        self.assertEqual(message.text, "It's a body\N{HORIZONTAL ELLIPSIS}\n")
        self.assertEqual(
            message.html,
            """<div dir="ltr">It's a body\N{HORIZONTAL ELLIPSIS}</div>\n""",
        )
        self.assertIsNone(message.spam_detected)

    def test_inbound_s3_failure_message(self):
        """Issue a helpful error when S3 download fails"""
        # Boto's error:
        # "An error occurred (403) when calling the HeadObject operation: Forbidden"
        from botocore.exceptions import ClientError

        self.mock_s3.download_fileobj.side_effect = ClientError(
            {"Error": {"Code": 403, "Message": "Forbidden"}},
            operation_name="HeadObject",
        )

        raw_ses_event = {
            "notificationType": "Received",
            "receipt": {
                "action": {
                    "type": "S3",
                    "bucketName": "YourBucket",
                    "objectKey": "inbound/the_object_key",
                }
            },
        }
        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "8f6dee70-c885-558a-be7d-bd48bbf5335e",
            "TopicArn": "arn:aws:sns:us-east-1:111111111111:SES_Inbound",
            "Message": json.dumps(raw_ses_event),
        }
        with self.assertRaisesMessage(
            AnymailAPIError,
            "Anymail AmazonSESInboundWebhookView couldn't download"
            " S3 object 'YourBucket:inbound/the_object_key'",
        ) as cm:
            self.post_from_sns("/anymail/amazon_ses/inbound/", raw_sns_message)
        # both Boto and Anymail exception class:
        self.assertIsInstance(cm.exception, ClientError)
        # original Boto message included:
        self.assertIn(
            "ClientError: An error occurred (403) when calling"
            " the HeadObject operation: Forbidden",
            str(cm.exception),
        )

    def test_incorrect_tracking_event(self):
        """The inbound webhook should warn if it receives tracking events"""
        raw_sns_message = {
            "Type": "Notification",
            "MessageId": "8f6dee70-c885-558a-be7d-bd48bbf5335e",
            "TopicArn": "arn:...:111111111111:SES_Tracking",
            "Message": '{"notificationType": "Delivery"}',
        }

        with self.assertRaisesMessage(
            AnymailConfigurationError,
            "You seem to have set an Amazon SES *sending* event or notification"
            " to publish to an SNS Topic that posts to Anymail's *inbound* webhook URL."
            " (SNS TopicArn arn:...:111111111111:SES_Tracking)",
        ):
            self.post_from_sns("/anymail/amazon_ses/inbound/", raw_sns_message)



import os
import unittest
from datetime import datetime, timedelta
from email.utils import formataddr

from django.test import SimpleTestCase, override_settings, tag

from anymail.exceptions import AnymailAPIError
from anymail.message import AnymailMessage

from .utils import AnymailTestMixin, sample_image_path

ANYMAIL_TEST_SENDGRID_API_KEY = os.getenv("ANYMAIL_TEST_SENDGRID_API_KEY")
ANYMAIL_TEST_SENDGRID_TEMPLATE_ID = os.getenv("ANYMAIL_TEST_SENDGRID_TEMPLATE_ID")
ANYMAIL_TEST_SENDGRID_DOMAIN = os.getenv("ANYMAIL_TEST_SENDGRID_DOMAIN")


@tag("sendgrid", "live")
@unittest.skipUnless(
    ANYMAIL_TEST_SENDGRID_API_KEY and ANYMAIL_TEST_SENDGRID_DOMAIN,
    "Set ANYMAIL_TEST_SENDGRID_API_KEY and ANYMAIL_TEST_SENDGRID_DOMAIN "
    "environment variables to run SendGrid integration tests",
)
@override_settings(
    ANYMAIL_SENDGRID_API_KEY=ANYMAIL_TEST_SENDGRID_API_KEY,
    ANYMAIL_SENDGRID_SEND_DEFAULTS={
        "esp_extra": {
            "mail_settings": {"sandbox_mode": {"enable": True}},
        }
    },
    EMAIL_BACKEND="anymail.backends.sendgrid.EmailBackend",
)
class SendGridBackendIntegrationTests(AnymailTestMixin, SimpleTestCase):
    """
    SendGrid v3 API integration tests

    These tests run against the **live** SendGrid API, using the
    environment variable `ANYMAIL_TEST_SENDGRID_API_KEY` as the API key
    If those variables are not set, these tests won't run.

    The SEND_DEFAULTS above force SendGrid's v3 sandbox mode, which avoids sending mail.
    (Sandbox sends also don't show in the activity feed, so disable that for live
    debugging.)

    The tests also use SendGrid's "sink domain" @sink.sendgrid.net for recipient
    addresses.
    https://support.sendgrid.com/hc/en-us/articles/201995663-Safely-Test-Your-Sending-Speed
    """

    def setUp(self):
        super().setUp()
        self.from_email = "from@%s" % ANYMAIL_TEST_SENDGRID_DOMAIN
        self.message = AnymailMessage(
            "Anymail SendGrid integration test",
            "Text content",
            self.from_email,
            ["to@sink.sendgrid.net"],
        )
        self.message.attach_alternative("<p>HTML content</p>", "text/html")

    def test_simple_send(self):
        # Example of getting the SendGrid send status and message id from the message
        sent_count = self.message.send()
        self.assertEqual(sent_count, 1)

        anymail_status = self.message.anymail_status
        sent_status = anymail_status.recipients["to@sink.sendgrid.net"].status
        message_id = anymail_status.recipients["to@sink.sendgrid.net"].message_id

        self.assertEqual(sent_status, "queued")  # SendGrid always queues
        self.assertUUIDIsValid(message_id)  # Anymail generates a UUID tracking id
        # set of all recipient statuses:
        self.assertEqual(anymail_status.status, {sent_status})
        self.assertEqual(anymail_status.message_id, message_id)

    def test_all_options(self):
        send_at = datetime.now().replace(microsecond=0) + timedelta(minutes=2)
        message = AnymailMessage(
            subject="Anymail all-options integration test",
            body="This is the text body",
            from_email=formataddr(("Test From, with comma", self.from_email)),
            to=["to1@sink.sendgrid.net", '"Recipient 2, OK?" <to2@sink.sendgrid.net>'],
            cc=["cc1@sink.sendgrid.net", "Copy 2 <cc2@sink.sendgrid.net>"],
            bcc=["bcc1@sink.sendgrid.net", "Blind Copy 2 <bcc2@sink.sendgrid.net>"],
            reply_to=['"Reply, with comma" <reply@example.com>', "reply2@example.com"],
            headers={"X-Anymail-Test": "value", "X-Anymail-Count": 3},
            metadata={"meta1": "simple string", "meta2": 2},
            send_at=send_at,
            tags=["tag 1", "tag 2"],
            track_clicks=True,
            track_opens=True,
            # this breaks activity feed if you don't have an asm group:
            # esp_extra={'asm': {'group_id': 1}},
        )
        message.attach("attachment1.txt", "Here is some\ntext for you", "text/plain")
        message.attach("attachment2.csv", "ID,Name\n1,Amy Lina", "text/csv")
        cid = message.attach_inline_image_file(sample_image_path())
        message.attach_alternative(
            "<p><b>HTML:</b> with <a href='http://example.com'>link</a>"
            "and image: <img src='cid:%s'></div>" % cid,
            "text/html",
        )

        message.send()
        # SendGrid always queues:
        self.assertEqual(message.anymail_status.status, {"queued"})

    def test_merge_data(self):
        message = AnymailMessage(
            subject="Anymail merge_data test: %field%",
            body="This body includes merge data: %field%",
            from_email=formataddr(("Test From", self.from_email)),
            to=["to1@sink.sendgrid.net", "Recipient 2 <to2@sink.sendgrid.net>"],
            reply_to=['"Merge data in reply name: %field%" <reply@example.com>'],
            merge_data={
                "to1@sink.sendgrid.net": {"field": "one"},
                "to2@sink.sendgrid.net": {"field": "two"},
            },
            esp_extra={
                "merge_field_format": "%{}%",
            },
            metadata={"meta1": "simple string", "meta2": 2},
            merge_metadata={
                "to1@sink.sendgrid.net": {"meta3": "recipient 1"},
                "to2@sink.sendgrid.net": {"meta3": "recipient 2"},
            },
            headers={
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
                "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
            },
            merge_headers={
                "to1@sink.sendgrid.net": {
                    "List-Unsubscribe": "<https://example.com/a/>",
                },
                "to2@sink.sendgrid.net": {
                    "List-Unsubscribe": "<https://example.com/b/>",
                },
            },
        )
        message.send()
        recipient_status = message.anymail_status.recipients
        self.assertEqual(recipient_status["to1@sink.sendgrid.net"].status, "queued")
        self.assertEqual(recipient_status["to2@sink.sendgrid.net"].status, "queued")

    @unittest.skipUnless(
        ANYMAIL_TEST_SENDGRID_TEMPLATE_ID,
        "Set the ANYMAIL_TEST_SENDGRID_TEMPLATE_ID environment variable "
        "to a template in your SendGrid account to test stored templates",
    )
    def test_stored_template(self):
        message = AnymailMessage(
            from_email=formataddr(("Test From", self.from_email)),
            to=["to@sink.sendgrid.net"],
            # Anymail's live test template has
            # merge fields "name", "order_no", and "dept"...
            template_id=ANYMAIL_TEST_SENDGRID_TEMPLATE_ID,
            merge_data={
                "to@sink.sendgrid.net": {
                    "name": "Test Recipient",
                    "order_no": "12345",
                },
            },
            merge_global_data={"dept": "Fulfillment"},
        )
        message.send()
        self.assertEqual(message.anymail_status.status, {"queued"})

    @override_settings(ANYMAIL_SENDGRID_API_KEY="Hey, that's not an API key!")
    def test_invalid_api_key(self):
        with self.assertRaises(AnymailAPIError) as cm:
            self.message.send()
        err = cm.exception
        self.assertEqual(err.status_code, 401)
        # Make sure the exception message includes SendGrid's response:
        self.assertIn("authorization grant is invalid", str(err))



import json
from base64 import b64encode
from datetime import date, datetime
from decimal import Decimal
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from unittest.mock import patch

from django.core import mail
from django.test import SimpleTestCase, override_settings, tag
from django.utils.timezone import (
    get_fixed_timezone,
    override as override_current_timezone,
)

from anymail.exceptions import (
    AnymailAPIError,
    AnymailConfigurationError,
    AnymailRecipientsRefused,
    AnymailSerializationError,
    AnymailUnsupportedFeature,
)
from anymail.message import AnymailMessage, attach_inline_image_file

from .mock_requests_backend import (
    RequestsBackendMockAPITestCase,
    SessionSharingTestCases,
)
from .utils import (
    SAMPLE_IMAGE_FILENAME,
    AnymailTestMixin,
    sample_image_content,
    sample_image_path,
)


@tag("unisender_go")
@override_settings(
    EMAIL_BACKEND="anymail.backends.unisender_go.EmailBackend",
    ANYMAIL={
        "UNISENDER_GO_API_KEY": "test_api_key",
        "UNISENDER_GO_API_URL": "https://go1.unisender.ru/ru/transactional/api/v1",
    },
)
class UnisenderGoBackendMockAPITestCase(RequestsBackendMockAPITestCase):
    DEFAULT_RAW_RESPONSE = json.dumps(
        {
            "status": "success",
            "job_id": "1rctPx-00021H-CcC4",
            "emails": ["to@example.com"],
        }
    ).encode("utf-8")
    DEFAULT_STATUS_CODE = 200
    DEFAULT_CONTENT_TYPE = "application/json"

    def setUp(self):
        super().setUp()

        # Patch uuid4 to generate predictable message_ids for testing
        patch_uuid4 = patch(
            "anymail.backends.unisender_go.uuid.uuid4",
            side_effect=[f"mocked-uuid-{n:d}" for n in range(1, 10)],
        )
        patch_uuid4.start()
        self.addCleanup(patch_uuid4.stop)

        # Simple message useful for many tests
        self.message = AnymailMessage(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )

    def set_mock_response(
        self, success_emails=None, failed_emails=None, job_id=None, **kwargs
    ):
        """
        Pass success_emails and/or failure_emails to generate an appropriate
        API response for those specific emails. Otherwise, arguments are as
        for super call.
        :param success_emails {list[str]}: addr-specs of emails that were delivered
        :param failure_emails {dict[str,str]}: mapping of addr-spec -> failure reason
        :param job_id {str}: optional specific job_id for response
        """
        if success_emails or failed_emails:
            assert "raw" not in kwargs
            assert "json_response" not in kwargs
            assert kwargs.get("status_code", 200) == 200
            kwargs["status_code"] = 200
            kwargs["json_data"] = {
                "status": "success",
                "job_id": job_id or "1rctPx-00021H-CcC4",
                "emails": success_emails or [],
            }
            if failed_emails:
                kwargs["json_data"]["failed_emails"] = failed_emails

        return super().set_mock_response(**kwargs)


@tag("unisender_go")
class UnisenderGoBackendStandardEmailTests(UnisenderGoBackendMockAPITestCase):
    """Test backend support for Django standard email features"""

    def test_send_mail(self):
        """Test basic API for simple send"""
        mail.send_mail(
            "Subject here",
            "Here is the message.",
            "from@sender.example.com",
            ["to@example.com"],
            fail_silently=False,
        )
        self.assert_esp_called(
            "https://go1.unisender.ru/ru/transactional/api/v1/email/send.json"
        )
        http_headers = self.get_api_call_headers()
        self.assertEqual(http_headers["X-API-KEY"], "test_api_key")
        self.assertEqual(http_headers["Accept"], "application/json")
        self.assertEqual(http_headers["Content-Type"], "application/json")

        data = self.get_api_call_json()
        self.assertEqual(data["message"]["subject"], "Subject here")
        self.assertEqual(data["message"]["body"], {"plaintext": "Here is the message."})
        self.assertEqual(data["message"]["from_email"], "from@sender.example.com")
        self.assertEqual(
            data["message"]["recipients"],
            [
                {
                    "email": "to@example.com",
                    # make sure the backend assigned the message_id
                    # for event tracking and notification
                    "metadata": {"anymail_id": "mocked-uuid-1"},
                }
            ],
        )

    def test_name_addr(self):
        """Make sure RFC2822 name-addr format (with display-name) is allowed

        (Test both sender and recipient addresses)
        """
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "From Name <from@example.com>",
            ["Recipient #1 <to1@example.com>", "to2@example.com"],
            cc=["Carbon Copy <cc1@example.com>", "cc2@example.com"],
            bcc=["Blind Copy <bcc1@example.com>", "bcc2@example.com"],
        )
        msg.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["from_email"], "from@example.com")
        self.assertEqual(data["message"]["from_name"], "From Name")

        recipients = data["message"]["recipients"]
        self.assertEqual(len(recipients), 6)
        self.assertEqual(recipients[0]["email"], "to1@example.com")
        self.assertEqual(recipients[0]["substitutions"]["to_name"], "Recipient #1")
        self.assertEqual(recipients[1]["email"], "to2@example.com")
        self.assertNotIn("substitutions", recipients[1])  # to_name not needed
        self.assertEqual(recipients[2]["email"], "cc1@example.com")
        self.assertEqual(recipients[2]["substitutions"]["to_name"], "Carbon Copy")
        self.assertEqual(recipients[3]["email"], "cc2@example.com")
        self.assertNotIn("substitutions", recipients[3])  # to_name not needed
        self.assertEqual(recipients[4]["email"], "bcc1@example.com")
        self.assertEqual(recipients[4]["substitutions"]["to_name"], "Blind Copy")
        self.assertEqual(recipients[5]["email"], "bcc2@example.com")
        self.assertNotIn("substitutions", recipients[5])  # to_name not needed

        # This also covers Unisender Go's special handling for cc/bcc
        headers = data["message"]["headers"]
        self.assertEqual(
            headers["to"], "Recipient #1 <to1@example.com>, to2@example.com"
        )
        self.assertEqual(
            headers["cc"], "Carbon Copy <cc1@example.com>, cc2@example.com"
        )
        self.assertNotIn("bcc", headers)

    def test_display_names_with_special_chars(self):
        # Verify workaround for Unisender Go bug parsing to/cc headers
        # with display names containing commas, angle brackets, or at sign
        self.message.to = [
            '"With, Comma" <to1@example.com>',
            '"angle <brackets>" <to2@example.com>',
            '"(without) special / chars" <to3@example.com>',
        ]
        self.message.cc = [
            '"Someone @example.com" <cc1@example.com>',
            '"[without] special & chars" <cc2@example.com>',
        ]
        self.message.send()
        data = self.get_api_call_json()
        headers = data["message"]["headers"]
        # display-name with , < > @ converted to RFC 2047 encoded word;
        # not necessary for display names with other special characters
        self.assertEqual(
            headers["to"],
            "=?utf-8?q?With=2C_Comma?= <to1@example.com>, "
            "=?utf-8?q?angle_=3Cbrackets=3E?= <to2@example.com>, "
            '"(without) special / chars" <to3@example.com>',
        )
        self.assertEqual(
            headers["cc"],
            "=?utf-8?q?Someone_=40example=2Ecom?= <cc1@example.com>, "
            '"[without] special & chars" <cc2@example.com>',
        )

    def test_html_message(self):
        text_content = "This is an important message."
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMultiAlternatives(
            "Subject", text_content, "from@example.com", ["to@example.com"]
        )
        email.attach_alternative(html_content, "text/html")
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["message"]["body"], {"plaintext": text_content, "html": html_content}
        )
        # Don't accidentally send the html part as an attachment:
        self.assertNotIn("attachments", data["message"])

    def test_html_only_message(self):
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMessage(
            "Subject", html_content, "from@example.com", ["to@example.com"]
        )
        email.content_subtype = "html"  # Main content is now text/html
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["body"], {"html": html_content})

    def test_amp_html_alternative(self):
        # Unisender Go *does* support text/x-amp-html alongside text/html
        self.message.attach_alternative("<p>HTML</p>", "text/html")
        self.message.attach_alternative("<p>And AMP HTML</p>", "text/x-amp-html")
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["body"]["html"], "<p>HTML</p>")
        self.assertEqual(data["message"]["body"]["amp"], "<p>And AMP HTML</p>")

    def test_extra_headers(self):
        self.message.extra_headers = {
            "X-Custom": "string",
            "X-Num": 123,
            "Reply-To": "noreply@example.com",
        }
        self.message.send()
        data = self.get_api_call_json()
        headers = data["message"]["headers"]
        self.assertEqual(headers["X-Custom"], "string")
        self.assertEqual(headers["X-Num"], 123)

        # Reply-To must be moved to separate param
        self.assertNotIn("Reply-To", headers)
        self.assertEqual(data["message"]["reply_to"], "noreply@example.com")
        self.assertNotIn("reply_to_name", data["message"])

    def test_extra_headers_serialization_error(self):
        self.message.extra_headers = {"X-Custom": Decimal(12.5)}
        with self.assertRaisesMessage(AnymailSerializationError, "Decimal"):
            self.message.send()

    def test_reply_to(self):
        # Unisender Go supports only a single reply-to
        self.message.reply_to = ['"Reply recipient" <reply@example.com']
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["reply_to"], "reply@example.com")
        self.assertEqual(data["message"]["reply_to_name"], "Reply recipient")

    def test_reply_to_name_workaround(self):
        # Check workaround for reply-to display-name containing special chars
        self.message.reply_to = ['"Reply (parens)" <reply@example.com']
        self.message.send()
        data = self.get_api_call_json()
        # Special chars force RFC 2047 encoded word
        self.assertEqual(
            data["message"]["reply_to_name"], "=?utf-8?q?Reply_=28parens=29?="
        )

    def test_attachments(self):
        text_content = "* Item one\n* Item two\n* Item three"
        self.message.attach(
            filename="test.txt", content=text_content, mimetype="text/plain"
        )

        # Should guess mimetype if not provided...
        png_content = b"PNG\xb4 pretend this is the contents of a png file"
        self.message.attach(filename="test.png", content=png_content)

        # Should work with a MIMEBase object (also tests no filename)...
        pdf_content = b"PDF\xb4 pretend this is valid pdf data"
        mimeattachment = MIMEBase("application", "pdf")
        mimeattachment.set_payload(pdf_content)
        self.message.attach(mimeattachment)

        self.message.send()
        data = self.get_api_call_json()
        attachments = data["message"]["attachments"]
        self.assertEqual(len(attachments), 3)

        self.assertEqual(
            attachments[0],
            {
                "name": "test.txt",
                "content": b64encode(text_content.encode("utf-8")).decode("ascii"),
                "type": "text/plain",
            },
        )
        self.assertEqual(
            attachments[1],
            {
                "name": "test.png",
                "content": b64encode(png_content).decode("ascii"),
                "type": "image/png",  # (type inferred from filename)
            },
        )
        self.assertEqual(
            attachments[2],
            {
                "name": "",  # no filename -- but param is required
                "content": b64encode(pdf_content).decode("ascii"),
                "type": "application/pdf",
            },
        )

    def test_embedded_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        cid = attach_inline_image_file(self.message, image_path)  # Read from a png file
        html_content = (
            '<p>This has an <img src="cid:%s" alt="inline" /> image.</p>' % cid
        )
        self.message.attach_alternative(html_content, "text/html")

        self.message.send()
        data = self.get_api_call_json()

        self.assertEqual(
            data["message"]["inline_attachments"],
            [
                {
                    "name": cid,
                    "content": b64encode(image_data).decode("ascii"),
                    "type": "image/png",  # (type inferred from filename)
                }
            ],
        )

    def test_attached_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        # option 1: attach as a file
        self.message.attach_file(image_path)

        # option 2: construct the MIMEImage and attach it directly
        image = MIMEImage(image_data)
        self.message.attach(image)

        self.message.send()

        image_data_b64 = b64encode(image_data).decode("ascii")
        data = self.get_api_call_json()
        self.assertEqual(
            data["message"]["attachments"][0],
            {
                "name": image_filename,  # the named one
                "content": image_data_b64,
                "type": "image/png",
            },
        )
        self.assertEqual(
            data["message"]["attachments"][1],
            {
                "name": "",  # the unnamed one
                "content": image_data_b64,
                "type": "image/png",
            },
        )

    def test_multiple_html_alternatives(self):
        # Multiple alternatives not allowed
        self.message.attach_alternative("<p>First html is OK</p>", "text/html")
        self.message.attach_alternative("<p>But not second html</p>", "text/html")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_non_html_alternative(self):
        # Only html alternatives allowed
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_api_failure(self):
        self.set_mock_response(status_code=400)
        with self.assertRaisesMessage(AnymailAPIError, "Unisender Go API response 400"):
            mail.send_mail("Subject", "Body", "from@example.com", ["to@example.com"])

        # Make sure fail_silently is respected
        self.set_mock_response(status_code=400)
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["to@example.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)

    def test_api_error_includes_details(self):
        """AnymailAPIError should include ESP's error message"""
        self.set_mock_response(
            status_code=400,
            json_data=[
                {
                    "status": "error",
                    "message": "Helpful explanation from Unisender Go",
                    "code": 999,
                },
            ],
        )
        with self.assertRaisesMessage(
            AnymailAPIError, "Helpful explanation from Unisender Go"
        ):
            self.message.send()


@tag("unisender_go")
class UnisenderGoBackendAnymailFeatureTests(UnisenderGoBackendMockAPITestCase):
    """Test backend support for Anymail added features"""

    def test_envelope_sender(self):
        # Unisender Go does not have a way to change envelope sender.
        self.message.envelope_sender = "anything@bounces.example.com"
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "envelope_sender"):
            self.message.send()

    def test_metadata(self):
        self.message.metadata = {"user_id": "12345", "items": 6, "float": 98.6}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["message"]["global_metadata"],
            {
                "user_id": "12345",
                "items": 6,
                "float": 98.6,
            },
        )

    def test_send_at(self):
        utc_plus_6 = get_fixed_timezone(6 * 60)
        utc_minus_8 = get_fixed_timezone(-8 * 60)

        with override_current_timezone(utc_plus_6):
            # Timezone-naive datetime assumed to be Django current_timezone
            self.message.send_at = datetime(2022, 10, 11, 12, 13, 14, 567)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["message"]["options"]["send_at"], "2022-10-11 06:13:14"
            )  # 12:13 UTC+6 == 06:13 UTC

            # Timezone-aware datetime converted to UTC:
            self.message.send_at = datetime(2016, 3, 4, 5, 6, 7, tzinfo=utc_minus_8)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["message"]["options"]["send_at"], "2016-03-04 13:06:07"
            )  # 05:06 UTC-8 == 13:06 UTC

            # Date-only treated as midnight in current timezone
            self.message.send_at = date(2022, 10, 22)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["message"]["options"]["send_at"], "2022-10-21 18:00:00"
            )  # 00:00 UTC+6 == 18:00-1d UTC

            # POSIX timestamp
            self.message.send_at = 1651820889  # 2022-05-06 07:08:09 UTC
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["message"]["options"]["send_at"], "2022-05-06 07:08:09"
            )

            # String passed unchanged (this is *not* portable between ESPs)
            self.message.send_at = "2013-11-12 01:02:03"
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["message"]["options"]["send_at"], "2013-11-12 01:02:03"
            )

    def test_tags(self):
        self.message.tags = ["receipt", "repeat-user"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertCountEqual(data["message"]["tags"], ["receipt", "repeat-user"])

    def test_tracking(self):
        # Test one way...
        self.message.track_clicks = False
        self.message.track_opens = True
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["track_links"], 0)
        self.assertEqual(data["message"]["track_read"], 1)

        # ...and the opposite way
        self.message.track_clicks = True
        self.message.track_opens = False
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["message"]["track_links"], 1)
        self.assertEqual(data["message"]["track_read"], 0)

    def test_template_id(self):
        self.message.template_id = "5997fcf6-2b9f-484d-acd5-7e9a99f0dc1f"
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["message"]["template_id"], "5997fcf6-2b9f-484d-acd5-7e9a99f0dc1f"
        )

    def test_merge_data(self):
        self.message.from_email = "from@example.com"
        self.message.to = [
            "alice@example.com",
            "Bob <bob@example.com>",
            "celia@example.com",
        ]
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Robert"},  # and leave group undefined
            # and no data for celia@example.com
        }
        self.message.merge_global_data = {
            "group": "Users",
            "site": "ExampleCo",
        }
        self.message.send()

        data = self.get_api_call_json()
        recipients = data["message"]["recipients"]
        self.assertEqual(recipients[0]["email"], "alice@example.com")
        self.assertEqual(
            recipients[0]["substitutions"], {"name": "Alice", "group": "Developers"}
        )
        self.assertEqual(recipients[1]["email"], "bob@example.com")
        self.assertEqual(
            # Make sure email display name (as "to_name") is combined with merge_data
            recipients[1]["substitutions"],
            {"name": "Robert", "to_name": "Bob"},
        )
        self.assertEqual(recipients[2]["email"], "celia@example.com")
        self.assertNotIn("substitutions", recipients[2])
        self.assertEqual(
            data["message"]["global_substitutions"],
            {"group": "Users", "site": "ExampleCo"},
        )

        # For batch send, must not include common "to" header
        headers = data["message"].get("headers", {})
        self.assertNotIn("to", headers)
        self.assertNotIn("cc", headers)

    def test_merge_metadata(self):
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123},
            "bob@example.com": {"order_id": 678, "tier": "premium"},
        }
        self.message.send()
        data = self.get_api_call_json()
        recipients = data["message"]["recipients"]
        # anymail_id added to other recipient metadata
        self.assertEqual(
            recipients[0]["metadata"],
            {
                "anymail_id": "mocked-uuid-1",
                "order_id": 123,
            },
        )
        self.assertEqual(
            recipients[1]["metadata"],
            {
                "anymail_id": "mocked-uuid-2",
                "order_id": 678,
                "tier": "premium",
            },
        )

        # For batch send, must not include common "to" header
        headers = data["message"].get("headers", {})
        self.assertNotIn("to", headers)
        self.assertNotIn("cc", headers)

    def test_merge_headers(self):
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.extra_headers = {
            "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
            "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
        }
        self.message.merge_headers = {
            "alice@example.com": {
                "List-Unsubscribe": "<https://example.com/a/>",
            },
            "bob@example.com": {
                "List-Unsubscribe": "<https://example.com/b/>",
            },
        }
        self.message.send()
        data = self.get_api_call_json()
        headers = data["message"]["headers"]
        recipients = data["message"]["recipients"]

        self.assertEqual(headers["List-Unsubscribe-Post"], "List-Unsubscribe=One-Click")

        # merge_headers List-Unsubscribe is handled as substitution:
        self.assertEqual(headers["List-Unsubscribe"], "{{Header__List_Unsubscribe}}")
        self.assertEqual(
            recipients[0]["substitutions"],
            {"Header__List_Unsubscribe": "<https://example.com/a/>"},
        )
        self.assertEqual(
            recipients[1]["substitutions"],
            # Header substitutions merged with other substitutions:
            {"Header__List_Unsubscribe": "<https://example.com/b/>", "to_name": "Bob"},
        )

    def test_unsupported_merge_headers(self):
        # Unisender Go only allows substitutions in the List-Unsubscribe header
        self.message.merge_headers = {"to@example.com": {"X-Other": "not supported"}}
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature, "'X-Other' in merge_headers"
        ):
            self.message.send()

    def test_cc_unsupported_with_batch_send(self):
        self.message.merge_data = {}
        self.message.cc = ["cc@example.com"]
        with self.assertRaisesMessage(
            AnymailUnsupportedFeature,
            "cc with batch send (merge_data, merge_metadata, or merge_headers)",
        ):
            self.message.send()

    @override_settings(ANYMAIL_IGNORE_UNSUPPORTED_FEATURES=True)
    def test_ignore_unsupported_cc_with_batch_send(self):
        self.message.merge_data = {}
        self.message.cc = ["cc@example.com"]
        self.message.bcc = ["bcc@example.com"]
        self.message.send()
        self.assertEqual(self.message.anymail_status.status, {"queued"})
        data = self.get_api_call_json()
        # Unisender Go prohibits "cc" header without "to" header,
        # and we can't include a "to" header for batch send,
        # so make sure we've removed the "cc" header when ignoring unsupported cc
        headers = data["message"].get("headers", {})
        self.assertNotIn("cc", headers)
        self.assertNotIn("to", headers)

    @override_settings(ANYMAIL_UNISENDER_GO_GENERATE_MESSAGE_ID=False)
    def test_default_omits_options(self):
        """Make sure by default we don't send any ESP-specific options.

        Options not specified by the caller should be omitted entirely from
        the API call (*not* sent as False or empty). This ensures
        that your ESP account settings apply by default.
        """
        self.message.send()
        message_data = self.get_api_call_json()["message"]
        self.assertNotIn("attachments", message_data)
        self.assertNotIn("from_name", message_data)
        self.assertNotIn("global_substitutions", message_data)
        self.assertNotIn("global_metadata", message_data)
        self.assertNotIn("inline_attachments", message_data)
        self.assertNotIn("options", message_data)
        self.assertNotIn("reply_to", message_data)
        self.assertNotIn("reply_to_name", message_data)
        self.assertNotIn("tags", message_data)
        self.assertNotIn("template_id", message_data)
        self.assertNotIn("track_links", message_data)
        self.assertNotIn("track_read", message_data)

        for recipient_data in message_data["recipients"]:
            self.assertNotIn("metadata", recipient_data)
            self.assertNotIn("substitutions", recipient_data)

    def test_esp_extra(self):
        self.message.send_at = "2022-02-22 22:22:22"
        self.message.esp_extra = {
            "global_language": "en",
            "skip_unsubscribe": 1,
            "template_engine": "velocity",
            "options": {
                "unsubscribe_url": "https://example.com/unsubscribe?id={{user_id}}",
                "smtp_pool_id": "custom-smtp-pool",
            },
        }
        self.message.send()
        data = self.get_api_call_json()
        # merged from esp_extra:
        self.assertEqual(data["message"]["global_language"], "en")
        self.assertEqual(data["message"]["skip_unsubscribe"], 1)
        self.assertEqual(data["message"]["template_engine"], "velocity")
        self.assertEqual(
            data["message"]["options"],
            {  # deep merge
                "send_at": "2022-02-22 22:22:22",
                "unsubscribe_url": "https://example.com/unsubscribe?id={{user_id}}",
                "smtp_pool_id": "custom-smtp-pool",
            },
        )

    # noinspection PyUnresolvedReferences
    def test_send_attaches_anymail_status(self):
        """The anymail_status should be attached to the message when it is sent"""
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to@example.com"],
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"queued"})
        self.assertEqual(msg.anymail_status.message_id, "mocked-uuid-1")
        self.assertEqual(
            msg.anymail_status.recipients["to@example.com"].status, "queued"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to@example.com"].message_id, "mocked-uuid-1"
        )
        self.assertEqual(
            msg.anymail_status.esp_response.content, self.DEFAULT_RAW_RESPONSE
        )

    def test_batch_recipients_get_unique_message_ids(self):
        """In a batch send, each recipient should get a distinct message_id"""
        # Unisender Go *always* uses batch send; no need to force by setting merge_data.
        self.set_mock_response(success_emails=["to1@example.com", "to2@example.com"])
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com", "Someone Else <to2@example.com>"],
        )
        msg.send()
        self.assertEqual(
            msg.anymail_status.message_id, {"mocked-uuid-1", "mocked-uuid-2"}
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].message_id, "mocked-uuid-1"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to2@example.com"].message_id, "mocked-uuid-2"
        )

    def test_rejected_recipient_status(self):
        self.message.to = [
            "duplicate@example.com",
            "Again <duplicate@example.com>",
            "Duplicate@example.com",  # addresses are case-insensitive
            "bounce@example.com",
            "mailbox-full@example.com",
            "webmaster@localhost",
            "spam-report@example.com",
        ]
        self.set_mock_response(
            # Note "duplicate" email will appear in both success and failed lists
            # (because Unisender Go sends the first one, fails remaining duplicates)
            success_emails=["duplicate@example.com"],
            failed_emails={
                "duplicate@example.com": "duplicate",
                "Duplicate@example.com": "duplicate",
                "bounce@example.com": "permanent_unavailable",
                "mailbox-full@example.com": "temporary_unavailable",
                "webmaster@localhost": "invalid",
                "spam-report@example.com": "unsubscribed",
            },
        )
        self.message.send()
        recipient_status = self.message.anymail_status.recipients
        self.assertEqual(recipient_status["duplicate@example.com"].status, "queued")
        self.assertEqual(
            # duplicate uses _first_ message_id (because first instance will be sent)
            recipient_status["duplicate@example.com"].message_id,
            "mocked-uuid-1",
        )
        self.assertEqual(recipient_status["bounce@example.com"].status, "rejected")
        self.assertIsNone(recipient_status["bounce@example.com"].message_id)
        self.assertEqual(recipient_status["mailbox-full@example.com"].status, "failed")
        self.assertIsNone(recipient_status["mailbox-full@example.com"].message_id)
        self.assertEqual(recipient_status["webmaster@localhost"].status, "invalid")
        self.assertIsNone(recipient_status["webmaster@localhost"].message_id)
        self.assertEqual(recipient_status["spam-report@example.com"].status, "rejected")
        self.assertIsNone(recipient_status["spam-report@example.com"].message_id)

    @override_settings(ANYMAIL_UNISENDER_GO_GENERATE_MESSAGE_ID=False)
    def test_disable_generate_message_id(self):
        """
        When not generating per-recipient message_id,
        use Unisender Go's job_id for all recipients.
        """
        self.set_mock_response(
            success_emails=["to1@example.com", "to2@example.com"],
            job_id="123456-000HHH-CcCc",
        )
        self.message.to = ["to1@example.com", "to2@example.com"]
        self.message.send()
        self.assertEqual(self.message.anymail_status.message_id, "123456-000HHH-CcCc")
        recipient_status = self.message.anymail_status.recipients
        self.assertEqual(
            recipient_status["to1@example.com"].message_id, "123456-000HHH-CcCc"
        )
        self.assertEqual(
            recipient_status["to2@example.com"].message_id, "123456-000HHH-CcCc"
        )

    # noinspection PyUnresolvedReferences
    def test_send_failed_anymail_status(self):
        """If the send fails, anymail_status should contain initial values"""
        self.set_mock_response(status_code=500)
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertIsNone(self.message.anymail_status.esp_response)

    def test_json_serialization_errors(self):
        """Try to provide more information about non-json-serializable data"""
        self.message.metadata = {"total": Decimal("19.99")}
        with self.assertRaises(AnymailSerializationError) as cm:
            self.message.send()
        err = cm.exception
        self.assertIsInstance(err, TypeError)  # compatibility with json.dumps
        # our added context:
        self.assertIn("Don't know how to send this data to Unisender Go", str(err))
        # original message:
        self.assertRegex(str(err), r"Decimal.*is not JSON serializable")


@tag("unisender_go")
class UnisenderGoBackendRecipientsRefusedTests(UnisenderGoBackendMockAPITestCase):
    """
    Should raise AnymailRecipientsRefused when *all* recipients are rejected or invalid
    """

    def test_recipients_refused(self):
        self.message.to = ["invalid@localhost", "reject@example.com"]
        self.set_mock_response(
            failed_emails={
                "invalid@localhost": "invalid",
                "reject@example.com": "permanent_unavailable",
            }
        )
        with self.assertRaises(AnymailRecipientsRefused):
            self.message.send()

    def test_fail_silently(self):
        self.message.to = ["invalid@localhost", "reject@example.com"]
        self.set_mock_response(
            failed_emails={
                "invalid@localhost": "invalid",
                "reject@example.com": "permanent_unavailable",
            }
        )
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)

    def test_mixed_response(self):
        """If *any* recipients are valid or queued, no exception is raised"""
        self.message.to = [
            "invalid@localhost",
            "valid@example.com",
            "reject@example.com",
            "also.valid@example.com",
        ]
        self.set_mock_response(
            success_emails=["valid@example.com", "also.valid@example.com"],
            failed_emails={
                "invalid@localhost": "invalid",
                "reject@example.com": "permanent_unavailable",
            },
        )
        sent = self.message.send()
        # one message sent, successfully, to 2 of 4 recipients:
        self.assertEqual(sent, 1)
        status = self.message.anymail_status
        self.assertEqual(status.recipients["invalid@localhost"].status, "invalid")
        self.assertEqual(status.recipients["valid@example.com"].status, "queued")
        self.assertEqual(status.recipients["reject@example.com"].status, "rejected")
        self.assertEqual(status.recipients["also.valid@example.com"].status, "queued")

    @override_settings(ANYMAIL_IGNORE_RECIPIENT_STATUS=True)
    def test_settings_override(self):
        """No exception with ignore setting"""
        self.message.to = ["invalid@localhost", "reject@example.com"]
        self.set_mock_response(
            failed_emails={
                "invalid@localhost": "invalid",
                "reject@example.com": "permanent_unavailable",
            }
        )
        sent = self.message.send()
        self.assertEqual(sent, 1)  # refused message is included in sent count


@tag("unisender_go")
class UnisenderGoBackendSessionSharingTestCase(
    SessionSharingTestCases, UnisenderGoBackendMockAPITestCase
):
    """Requests session sharing tests"""

    pass  # tests are defined in SessionSharingTestCases


@tag("unisender_go")
@override_settings(EMAIL_BACKEND="anymail.backends.unisender_go.EmailBackend")
class UnisenderGoBackendImproperlyConfiguredTests(AnymailTestMixin, SimpleTestCase):
    """Test ESP backend without required settings in place"""

    def test_missing_auth(self):
        with self.assertRaisesRegex(
            AnymailConfigurationError, r"\bUNISENDER_GO_API_KEY\b"
        ):
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])



from __future__ import annotations

import json
import typing
from datetime import datetime, timezone
from hashlib import md5

from django.http import HttpRequest, HttpResponse
from django.utils.crypto import constant_time_compare

from ..exceptions import AnymailWebhookValidationFailure
from ..signals import AnymailTrackingEvent, EventType, RejectReason, tracking
from ..utils import get_anymail_setting
from .base import AnymailBaseWebhookView


class UnisenderGoTrackingWebhookView(AnymailBaseWebhookView):
    """Handler for Unisender Go delivery and engagement tracking webhooks"""

    # See https://godocs.unisender.ru/web-api-ref#callback-format for webhook payload

    esp_name = "Unisender Go"
    signal = tracking
    warn_if_no_basic_auth = False  # because we validate against signature

    api_key: str | None = None  # allows kwargs override

    event_types = {
        "sent": EventType.SENT,
        "delivered": EventType.DELIVERED,
        "opened": EventType.OPENED,
        "clicked": EventType.CLICKED,
        "unsubscribed": EventType.UNSUBSCRIBED,
        "subscribed": EventType.SUBSCRIBED,
        "spam": EventType.COMPLAINED,
        "soft_bounced": EventType.DEFERRED,
        "hard_bounced": EventType.BOUNCED,
    }

    reject_reasons = {
        "err_user_unknown": RejectReason.BOUNCED,
        "err_user_inactive": RejectReason.BOUNCED,
        "err_will_retry": None,  # not rejected
        "err_mailbox_discarded": RejectReason.BOUNCED,
        "err_mailbox_full": RejectReason.BOUNCED,
        "err_spam_rejected": RejectReason.SPAM,
        "err_blacklisted": RejectReason.BLOCKED,
        "err_too_large": RejectReason.BOUNCED,
        "err_unsubscribed": RejectReason.UNSUBSCRIBED,
        "err_unreachable": RejectReason.BOUNCED,
        "err_skip_letter": RejectReason.BOUNCED,
        "err_domain_inactive": RejectReason.BOUNCED,
        "err_destination_misconfigured": RejectReason.BOUNCED,
        "err_delivery_failed": RejectReason.OTHER,
        "err_spam_skipped": RejectReason.SPAM,
        "err_lost": RejectReason.OTHER,
    }

    http_method_names = ["post", "head", "options", "get"]

    def __init__(self, **kwargs):
        api_key = get_anymail_setting(
            "api_key", esp_name=self.esp_name, allow_bare=True, kwargs=kwargs
        )
        self.api_key_bytes = api_key.encode("ascii")
        super().__init__(**kwargs)

    def get(
        self, request: HttpRequest, *args: typing.Any, **kwargs: typing.Any
    ) -> HttpResponse:
        # Unisender Go verifies the webhook with a GET request at configuration time
        return HttpResponse()

    def parse_json_body(self, request: HttpRequest) -> dict | list | None:
        # Cache parsed JSON request.body on the request.
        if hasattr(request, "_parsed_json"):
            parsed = getattr(request, "_parsed_json")
        else:
            parsed = json.loads(request.body.decode())
            setattr(request, "_parsed_json", parsed)
        return parsed

    def validate_request(self, request: HttpRequest) -> None:
        """
        Validate Unisender Go webhook signature:
        "MD5 hash of the string body of the message, with the auth value replaced
        by the api_key of the user/project whose handler is being called."
        https://godocs.unisender.ru/web-api-ref#callback-format
        """
        # This must avoid any assumptions about how Unisender Go serializes JSON
        # (key order, spaces, Unicode encoding vs. \u escapes, etc.). But we do
        # assume the "auth" field MD5 hash is unique within the serialized JSON,
        # so that we can use string replacement to calculate the expected hash.
        body = request.body
        try:
            parsed = self.parse_json_body(request)
            actual_auth = parsed["auth"]
            actual_auth_bytes = actual_auth.encode()
        except (AttributeError, KeyError, ValueError):
            raise AnymailWebhookValidationFailure(
                "Unisender Go webhook called with invalid payload"
            )

        body_to_sign = body.replace(actual_auth_bytes, self.api_key_bytes)
        expected_auth = md5(body_to_sign).hexdigest()
        if not constant_time_compare(actual_auth, expected_auth):
            # If webhook has a selected project, include the project_id in the error.
            try:
                project_id = parsed["events_by_user"][0]["project_id"]
            except (KeyError, IndexError):
                project_id = parsed.get("project_id")  # try "single event" payload
            is_for_project = f" is for Project ID {project_id}" if project_id else ""
            raise AnymailWebhookValidationFailure(
                "Unisender Go webhook called with incorrect signature"
                f" (check Anymail UNISENDER_GO_API_KEY setting{is_for_project})"
            )

    def parse_events(self, request: HttpRequest) -> list[AnymailTrackingEvent]:
        parsed = self.parse_json_body(request)
        # Unisender Go has two options for webhook payloads. We support both.
        try:
            events_by_user = parsed["events_by_user"]
        except KeyError:
            # "Use single event": one flat dict, combining "event_data" fields
            # with "event_name", "user_id", "project_id", etc.
            if parsed["event_name"] == "transactional_email_status":
                esp_events = [parsed]
            else:
                esp_events = []
        else:
            # Not "use single event": we want the "event_data" from all events
            # with event_name "transactional_email_status".
            assert len(events_by_user) == 1  # "A single element array" per API docs
            esp_events = [
                event["event_data"]
                for event in events_by_user[0]["events"]
                if event["event_name"] == "transactional_email_status"
            ]

        return [self.esp_to_anymail_event(esp_event) for esp_event in esp_events]

    def esp_to_anymail_event(self, event_data: dict) -> AnymailTrackingEvent:
        event_type = self.event_types.get(event_data["status"], EventType.UNKNOWN)

        # Unisender Go does not provide any way to deduplicate webhook calls.
        # (There is an "ID" HTTP header, but it has a unique value for every
        # webhook call--including retransmissions of earlier failed calls.)
        event_id = None

        # event_time is ISO-like, without a stated time zone. (But it's UTC per docs.)
        try:
            timestamp = datetime.fromisoformat(event_data["event_time"]).replace(
                tzinfo=timezone.utc
            )
        except KeyError:
            timestamp = None

        # Extract our message_id (see backend UNISENDER_GO_GENERATE_MESSAGE_ID).
        metadata = event_data.get("metadata", {}).copy()
        message_id = metadata.pop("anymail_id", event_data.get("job_id"))

        delivery_info = event_data.get("delivery_info", {})
        delivery_status = delivery_info.get("delivery_status", "")
        if delivery_status.startswith("err"):
            reject_reason = self.reject_reasons.get(delivery_status, RejectReason.OTHER)
        else:
            reject_reason = None

        description = delivery_info.get("delivery_status") or event_data.get("comment")
        mta_response = delivery_info.get("destination_response")

        return AnymailTrackingEvent(
            event_type=event_type,
            timestamp=timestamp,
            message_id=message_id,
            event_id=event_id,
            recipient=event_data["email"],
            reject_reason=reject_reason,
            description=description,
            mta_response=mta_response,
            metadata=metadata,
            click_url=event_data.get("url"),
            user_agent=delivery_info.get("user_agent"),
            esp_event=event_data,
        )
